import py
from rpython.jit.backend.llsupport.descr import get_size_descr,\
     get_field_descr, get_array_descr, ArrayDescr, FieldDescr,\
     SizeDescr, get_interiorfield_descr
from rpython.jit.backend.llsupport.gc import GcLLDescr_boehm,\
     GcLLDescr_framework
from rpython.jit.backend.llsupport import jitframe
from rpython.jit.metainterp.gc import get_description
from rpython.jit.tool.oparser import parse
from rpython.jit.metainterp.optimizeopt.util import equaloplists
from rpython.jit.metainterp.history import JitCellToken, FLOAT
from rpython.jit.metainterp.history import AbstractFailDescr
from rpython.rtyper.lltypesystem import lltype, llmemory, rffi
from rpython.rtyper import rclass
from rpython.jit.backend.llsupport.symbolic import (WORD,
        get_array_token)

class Evaluator(object):
    def __init__(self, scope):
        self.scope = scope
    def __getitem__(self, key):
        return eval(key, self.scope)


class FakeLoopToken(object):
    pass

o_vtable = lltype.malloc(rclass.OBJECT_VTABLE, immortal=True)

class RewriteTests(object):
    def check_rewrite(self, frm_operations, to_operations, **namespace):
        def setfield(baseptr, newvalue, descr):
            assert isinstance(baseptr, str)
            assert isinstance(newvalue, (str, int))
            assert not isinstance(descr, (str, int))
            return 'gc_store(%s, %d, %s, %d)' % (baseptr, descr.offset,
                                                 newvalue, descr.field_size)
        def zero_array(baseptr, start, length, descr_name, descr):
            assert isinstance(baseptr, str)
            assert isinstance(start, (str, int))
            assert isinstance(length, (str, int))
            assert isinstance(descr_name, str)
            assert not isinstance(descr, (str,int))
            itemsize = descr.itemsize
            start = start * itemsize
            length_scale = 1
            if isinstance(length, str):
                length_scale = itemsize
            else:
                length = length * itemsize
            return 'zero_array(%s, %s, %s, 1, %d, descr=%s)' % \
                      (baseptr, start, length, length_scale, descr_name)
        def setarrayitem(baseptr, index, newvalue, descr):
            assert isinstance(baseptr, str)
            assert isinstance(index, (str, int))
            assert isinstance(newvalue, (str, int))
            assert not isinstance(descr, (str, int))
            if isinstance(index, int):
                offset = descr.basesize + index * descr.itemsize
                return 'gc_store(%s, %d, %s, %d)' % (baseptr, offset,
                                                     newvalue, descr.itemsize)
            else:
                return 'gc_store_indexed(%s, %s, %s, %d, %d, %s)' % (
                    baseptr, index, newvalue,
                    descr.itemsize, descr.basesize, descr.itemsize)
        #
        WORD = globals()['WORD']
        S = lltype.GcStruct('S', ('x', lltype.Signed),
                                 ('y', lltype.Signed))
        sdescr = get_size_descr(self.gc_ll_descr, S)
        sdescr.tid = 1234
        #
        T = lltype.GcStruct('T', ('y', lltype.Signed),
                                 ('z', lltype.Ptr(S)),
                                 ('t', lltype.Signed))
        tdescr = get_size_descr(self.gc_ll_descr, T)
        tdescr.tid = 5678
        tzdescr = get_field_descr(self.gc_ll_descr, T, 'z')
        myT = lltype.cast_opaque_ptr(llmemory.GCREF,
                                     lltype.malloc(T, zero=True))
        self.myT = myT
        #
        A = lltype.GcArray(lltype.Signed)
        adescr = get_array_descr(self.gc_ll_descr, A)
        adescr.tid = 4321
        alendescr = adescr.lendescr
        #
        B = lltype.GcArray(lltype.Char)
        bdescr = get_array_descr(self.gc_ll_descr, B)
        bdescr.tid = 8765
        blendescr = bdescr.lendescr
        #
        C = lltype.GcArray(lltype.Ptr(S))
        cdescr = get_array_descr(self.gc_ll_descr, C)
        cdescr.tid = 8111
        clendescr = cdescr.lendescr
        #
        S1 = lltype.GcStruct('S1')
        S1I = lltype.GcArray(('x', lltype.Ptr(S1)),
                             ('y', lltype.Ptr(S1)),
                             ('z', lltype.Ptr(S1)))
        itzdescr = get_interiorfield_descr(self.gc_ll_descr, S1I, 'z')
        itydescr = get_interiorfield_descr(self.gc_ll_descr, S1I, 'y')
        itxdescr = get_interiorfield_descr(self.gc_ll_descr, S1I, 'x')
        S2I = lltype.GcArray(('x', lltype.Ptr(S1)),
                             ('y', lltype.Ptr(S1)),
                             ('z', lltype.Ptr(S1)),
                             ('t', lltype.Ptr(S1)))   # size is a power of two
        s2i_item_size_in_bits = (4 if WORD == 4 else 5)
        ity2descr = get_interiorfield_descr(self.gc_ll_descr, S2I, 'y')
        R1 = lltype.GcStruct('R', ('x', lltype.Signed),
                                  ('y', lltype.Float),
                                  ('z', lltype.Ptr(S1)))
        xdescr = get_field_descr(self.gc_ll_descr, R1, 'x')
        ydescr = get_field_descr(self.gc_ll_descr, R1, 'y')
        zdescr = get_field_descr(self.gc_ll_descr, R1, 'z')
        myR1 = lltype.cast_opaque_ptr(llmemory.GCREF,
                                      lltype.malloc(R1, zero=True))
        myR1b = lltype.cast_opaque_ptr(llmemory.GCREF,
                                       lltype.malloc(R1, zero=True))
        self.myR1 = myR1
        self.myR1b = myR1b
        #
        E = lltype.GcStruct('Empty')
        edescr = get_size_descr(self.gc_ll_descr, E)
        edescr.tid = 9000
        #
        vtable_descr = self.gc_ll_descr.fielddescr_vtable
        O = lltype.GcStruct('O', ('parent', rclass.OBJECT),
                                 ('x', lltype.Signed))
        o_descr = self.cpu.sizeof(O, True)
        o_vtable = globals()['o_vtable']
        #
        tiddescr = self.gc_ll_descr.fielddescr_tid
        wbdescr = self.gc_ll_descr.write_barrier_descr
        #
        F = lltype.GcArray(lltype.Float)
        fdescr = get_array_descr(self.gc_ll_descr, F)
        SF = lltype.GcArray(lltype.SingleFloat)
        sfdescr = get_array_descr(self.gc_ll_descr, SF)
        RAW_SF = lltype.Array(lltype.SingleFloat)
        raw_sfdescr = get_array_descr(self.gc_ll_descr, RAW_SF)
        #
        strdescr     = self.gc_ll_descr.str_descr
        str_basesize = self.gc_ll_descr.str_descr.basesize - 1
        unicodedescr = self.gc_ll_descr.unicode_descr
        strlendescr     = strdescr.lendescr
        unicodelendescr = unicodedescr.lendescr
        strhashdescr     = self.gc_ll_descr.str_hash_descr
        unicodehashdescr = self.gc_ll_descr.unicode_hash_descr
        uni_basesize  = unicodedescr.basesize
        uni_itemscale = {2: 1, 4: 2}[unicodedescr.itemsize]
        memcpy_fn = self.gc_ll_descr.memcpy_fn
        memcpy_descr = self.gc_ll_descr.memcpy_descr

        casmdescr = JitCellToken()
        clt = FakeLoopToken()
        clt._ll_initial_locs = [0, 8]
        frame_info = lltype.malloc(jitframe.JITFRAMEINFO, flavor='raw')
        clt.frame_info = frame_info
        frame_info.jfi_frame_depth = 13
        frame_info.jfi_frame_size = 255
        framedescrs = self.gc_ll_descr.getframedescrs(self.cpu)
        framelendescr = framedescrs.arraydescr.lendescr
        jfi_frame_depth = framedescrs.jfi_frame_depth
        jfi_frame_size = framedescrs.jfi_frame_size
        jf_frame_info = framedescrs.jf_frame_info
        jf_savedata = framedescrs.jf_savedata
        jf_force_descr = framedescrs.jf_force_descr
        jf_descr = framedescrs.jf_descr
        jf_guard_exc = framedescrs.jf_guard_exc
        jf_forward = framedescrs.jf_forward
        signedframedescr = self.cpu.signedframedescr
        floatframedescr = self.cpu.floatframedescr
        casmdescr.compiled_loop_token = clt

        #
        guarddescr = AbstractFailDescr()
        #
        namespace.update(locals())
        #
        for funcname in self.gc_ll_descr._generated_functions:
            namespace[funcname] = self.gc_ll_descr.get_malloc_fn(funcname)
            namespace[funcname + '_descr'] = getattr(self.gc_ll_descr,
                                                     '%s_descr' % funcname)
        #
        ops = parse(frm_operations, namespace=namespace)
        expected = parse(to_operations % Evaluator(namespace),
                         namespace=namespace)
        self.gcrefs = []
        operations = self.gc_ll_descr.rewrite_assembler(self.cpu,
                                                        ops.operations,
                                                        self.gcrefs)
        remap = {}
        for a, b in zip(ops.inputargs, expected.inputargs):
            remap[b] = a
        equaloplists(operations, expected.operations, remap=remap)
        lltype.free(frame_info, flavor='raw')

class FakeTracker(object):
    pass

class BaseFakeCPU(object):
    JITFRAME_FIXED_SIZE = 0

    load_constant_offset = True
    load_supported_factors = (1,2,4,8)
    supports_load_effective_address = True

    translate_support_code = None

    def __init__(self):
        self.tracker = FakeTracker()
        self._cache = {}
        self.signedframedescr = ArrayDescr(3, 8, FieldDescr('len', 0, 0, 0), 0)
        self.floatframedescr = ArrayDescr(5, 8, FieldDescr('len', 0, 0, 0), 0)

    def getarraydescr_for_frame(self, tp):
        if tp == FLOAT:
            return self.floatframedescr
        return self.signedframedescr

    def unpack_arraydescr_size(self, d):
        return 0, d.itemsize, 0

    def unpack_fielddescr(self, d):
        return d.offset

    def arraydescrof(self, ARRAY):
        try:
            return self._cache[ARRAY]
        except KeyError:
            r = ArrayDescr(1, 2, FieldDescr('len', 0, 0, 0), 0)
            self._cache[ARRAY] = r
            return r

    def fielddescrof(self, STRUCT, fname):
        key = (STRUCT, fname)
        try:
            return self._cache[key]
        except KeyError:
            r = FieldDescr(fname, 1, 1, 1)
            self._cache[key] = r
            return r

    def cast_adr_to_int(self, adr):
        return llmemory.AddressAsInt(adr)

class TestBoehm(RewriteTests):
    def setup_method(self, meth):
        class FakeCPU(BaseFakeCPU):
            def sizeof(self, STRUCT, is_object):
                assert is_object
                return SizeDescr(102, gc_fielddescrs=[],
                                 vtable=o_vtable)
        self.cpu = FakeCPU()
        self.gc_ll_descr = GcLLDescr_boehm(None, None, None)

    def test_new(self):
        self.check_rewrite("""
            []
            p0 = new(descr=sdescr)
            jump()
        """, """
            [p1]
            p0 = call_r(ConstClass(malloc_fixedsize), %(sdescr.size)d,\
                        descr=malloc_fixedsize_descr)
            check_memory_error(p0)
            jump()
        """)

    def test_no_collapsing(self):
        self.check_rewrite("""
            []
            p0 = new(descr=sdescr)
            p1 = new(descr=sdescr)
            jump()
        """, """
            []
            p0 = call_r(ConstClass(malloc_fixedsize), %(sdescr.size)d,\
                        descr=malloc_fixedsize_descr)
            check_memory_error(p0)
            p1 = call_r(ConstClass(malloc_fixedsize), %(sdescr.size)d,\
                        descr=malloc_fixedsize_descr)
            check_memory_error(p1)
            jump()
        """)

    def test_new_array_fixed(self):
        self.check_rewrite("""
            []
            p0 = new_array(10, descr=adescr)
            jump()
        """, """
            []
            p0 = call_r(ConstClass(malloc_array),           \
                                %(adescr.basesize)d,        \
                                10,                         \
                                %(adescr.itemsize)d,        \
                                %(adescr.lendescr.offset)d, \
                                descr=malloc_array_descr)
            check_memory_error(p0)
            jump()
        """)
##      should ideally be:
##            p0 = call_r(ConstClass(malloc_fixedsize), \
##                                %(adescr.basesize + 10 * adescr.itemsize)d, \
##                                descr=malloc_fixedsize_descr)
##            setfield_gc(p0, 10, descr=alendescr)

    def test_new_array_variable(self):
        self.check_rewrite("""
            [i1]
            p0 = new_array(i1, descr=adescr)
            jump()
        """, """
            [i1]
            p0 = call_r(ConstClass(malloc_array),   \
                                %(adescr.basesize)d,        \
                                i1,                         \
                                %(adescr.itemsize)d,        \
                                %(adescr.lendescr.offset)d, \
                                descr=malloc_array_descr)
            check_memory_error(p0)
            jump()
        """)

    def test_new_with_vtable(self):
        self.check_rewrite("""
            []
            p0 = new_with_vtable(descr=o_descr)
            jump()
        """, """
            [p1]
            p0 = call_r(ConstClass(malloc_fixedsize), 102, \
                                descr=malloc_fixedsize_descr)
            check_memory_error(p0)
            gc_store(p0, 0, ConstClass(o_vtable), %(vtable_descr.field_size)s)
            jump()
        """)

    def test_newstr(self):
        self.check_rewrite("""
            [i1]
            p0 = newstr(i1)
            jump()
        """, """
            [i1]
            p0 = call_r(ConstClass(malloc_array),         \
                                %(strdescr.basesize)d,    \
                                i1,                       \
                                %(strdescr.itemsize)d,    \
                                %(strlendescr.offset)d,   \
                                descr=malloc_array_descr)
            check_memory_error(p0)
            jump()
        """)

    def test_newunicode(self):
        self.check_rewrite("""
            [i1]
            p0 = newunicode(10)
            jump()
        """, """
            [i1]
            p0 = call_r(ConstClass(malloc_array),           \
                                %(unicodedescr.basesize)d,  \
                                10,                         \
                                %(unicodedescr.itemsize)d,  \
                                %(unicodelendescr.offset)d, \
                                descr=malloc_array_descr)
            check_memory_error(p0)
            jump()
        """)
##      should ideally be:
##            p0 = call_r(ConstClass(malloc_fixedsize),           \
##                                %(unicodedescr.basesize +       \
##                                  10 * unicodedescr.itemsize)d, \
##                                descr=malloc_fixedsize_descr)
##            setfield_gc(p0, 10, descr=unicodelendescr)


class TestFramework(RewriteTests):
    def setup_method(self, meth):
        class config_(object):
            class translation(object):
                gc = 'minimark'
                gcrootfinder = 'shadowstack'
                gctransformer = 'framework'
                gcremovetypeptr = False
        gcdescr = get_description(config_)
        self.gc_ll_descr = GcLLDescr_framework(gcdescr, None, None, None,
                                               really_not_translated=True)
        self.gc_ll_descr.write_barrier_descr.has_write_barrier_from_array = (
            lambda cpu: True)
        self.gc_ll_descr.malloc_zero_filled = False
        #
        class FakeCPU(BaseFakeCPU):
            def sizeof(self, STRUCT, is_object):
                descr = SizeDescr(104, gc_fielddescrs=[])
                descr.tid = 9315
                return descr
        self.cpu = FakeCPU()

    def test_rewrite_assembler_new_to_malloc(self):
        self.check_rewrite("""
            [p1]
            p0 = new(descr=sdescr)
            jump()
        """, """
            [p1]
            p0 = call_malloc_nursery(%(sdescr.size)d)
            gc_store(p0, 0, 1234, 8)
            jump()
        """)

    def test_rewrite_assembler_new3_to_malloc(self):
        self.check_rewrite("""
            []
            p0 = new(descr=sdescr)
            p1 = new(descr=tdescr)
            p2 = new(descr=sdescr)
            jump()
        """, """
            []
            p0 = call_malloc_nursery(   \
                               %(sdescr.size + tdescr.size + sdescr.size)d)
            gc_store(p0, 0, 1234, 8)
            p1 = nursery_ptr_increment(p0, %(sdescr.size)d)
            gc_store(p1, 0, 5678, 8)
            p2 = nursery_ptr_increment(p1, %(tdescr.size)d)
            gc_store(p2, 0, 1234, 8)
            %(setfield('p1', 0, tdescr.gc_fielddescrs[0]))s
            jump()
        """)

    def test_rewrite_assembler_new_array_fixed_to_malloc(self):
        self.check_rewrite("""
            []
            p0 = new_array(10, descr=adescr)
            jump()
        """, """
            []
            p0 = call_malloc_nursery(    \
                                %(adescr.basesize + 10 * adescr.itemsize)d)
            gc_store(p0, 0, 4321, %(tiddescr.field_size)s)
            gc_store(p0, 0, 10, %(alendescr.field_size)s)
            jump()
        """)

    def test_rewrite_assembler_new_and_new_array_fixed_to_malloc(self):
        self.check_rewrite("""
            []
            p0 = new(descr=sdescr)
            p1 = new_array(10, descr=adescr)
            jump()
        """, """
            []
            p0 = call_malloc_nursery(                                  \
                                %(sdescr.size +                        \
                                  adescr.basesize + 10 * adescr.itemsize)d)
            gc_store(p0, 0, 1234, %(tiddescr.field_size)s)
            p1 = nursery_ptr_increment(p0, %(sdescr.size)d)
            gc_store(p1, 0, 4321, %(tiddescr.field_size)s)
            gc_store(p1, 0, 10, %(alendescr.field_size)s)
            jump()
        """)

    def test_rewrite_assembler_round_up(self):
        self.check_rewrite("""
            []
            p0 = new_array(6, descr=bdescr)
            jump()
        """, """
            []
            p0 = call_malloc_nursery(%(bdescr.basesize + 8)d)
            gc_store(p0, 0, 8765, %(tiddescr.field_size)s)
            gc_store(p0, 0, 6, %(blendescr.field_size)s)
            jump()
        """)

    def test_rewrite_assembler_round_up_always(self):
        self.check_rewrite("""
            []
            p0 = new_array(5, descr=bdescr)
            p1 = new_array(5, descr=bdescr)
            p2 = new_array(5, descr=bdescr)
            p3 = new_array(5, descr=bdescr)
            jump()
        """, """
            []
            p0 = call_malloc_nursery(%(4 * (bdescr.basesize + 8))d)
            gc_store(p0, 0, 8765, %(tiddescr.field_size)s)
            gc_store(p0, 0, 5, %(blendescr.field_size)s)
            p1 = nursery_ptr_increment(p0, %(bdescr.basesize + 8)d)
            gc_store(p1, 0, 8765, %(tiddescr.field_size)s)
            gc_store(p1, 0, 5, %(blendescr.field_size)s)
            p2 = nursery_ptr_increment(p1, %(bdescr.basesize + 8)d)
            gc_store(p2, 0, 8765, %(tiddescr.field_size)s)
            gc_store(p2, 0, 5, %(blendescr.field_size)s)
            p3 = nursery_ptr_increment(p2, %(bdescr.basesize + 8)d)
            gc_store(p3, 0, 8765, %(tiddescr.field_size)s)
            gc_store(p3, 0, 5, %(blendescr.field_size)s)
            jump()
        """)

    def test_rewrite_assembler_minimal_size(self):
        self.check_rewrite("""
            []
            p0 = new(descr=edescr)
            p1 = new(descr=edescr)
            jump()
        """, """
            []
            p0 = call_malloc_nursery(%(4*WORD)d)
            gc_store(p0, 0,  9000, %(tiddescr.field_size)s)
            p1 = nursery_ptr_increment(p0, %(2*WORD)d)
            gc_store(p1, 0,  9000, %(tiddescr.field_size)s)
            jump()
        """)

    def test_rewrite_assembler_variable_size(self):
        self.check_rewrite("""
            [i0]
            p0 = new_array(i0, descr=bdescr)
            jump(i0)
        """, """
            [i0]
            p0 = call_malloc_nursery_varsize(0, 1, i0, descr=bdescr)
            gc_store(p0, 0, i0, %(bdescr.basesize)s)
            jump(i0)
        """)

    def test_rewrite_new_string(self):
        self.check_rewrite("""
        [i0]
        p0 = newstr(i0)
        jump(i0)
        """, """
        [i0]
        p0 = call_malloc_nursery_varsize(1, 1, i0, descr=strdescr)
        gc_store(p0, %(strlendescr.offset)s, i0, %(strlendescr.field_size)s)
        gc_store(p0, 0, 0, %(strlendescr.field_size)s)
        jump(i0)
        """)

    def test_rewrite_assembler_nonstandard_array(self):
        # a non-standard array is a bit hard to get; e.g. GcArray(Float)
        # is like that on Win32, but not on Linux.  Build one manually...
        NONSTD = lltype.GcArray(lltype.Float)
        nonstd_descr = get_array_descr(self.gc_ll_descr, NONSTD)
        nonstd_descr.tid = 6464
        nonstd_descr.basesize = 64      # <= hacked
        nonstd_descr.itemsize = 8
        nonstd_descr_gcref = 123
        # REVIEW: added descr=nonstd_descr to setarrayitem
        # is it even valid to have a setarrayitem WITHOUT a descr?
        self.check_rewrite("""
            [i0, p1]
            p0 = new_array(i0, descr=nonstd_descr)
            setarrayitem_gc(p0, i0, p1, descr=nonstd_descr)
            jump(i0)
        """, """
            [i0, p1]
            p0 = call_r(ConstClass(malloc_array_nonstandard),         \
                                64, 8,                                \
                                %(nonstd_descr.lendescr.offset)d,     \
                                6464, i0,                             \
                                descr=malloc_array_nonstandard_descr)
            check_memory_error(p0)
            cond_call_gc_wb_array(p0, i0, descr=wbdescr)
            gc_store_indexed(p0, i0, p1, 8, 64, 8)
            jump(i0)
        """, nonstd_descr=nonstd_descr)

    def test_rewrite_assembler_maximal_size_1(self):
        self.gc_ll_descr.max_size_of_young_obj = 100
        self.check_rewrite("""
            []
            p0 = new_array(103, descr=bdescr)
            jump()
        """, """
            []
            p0 = call_r(ConstClass(malloc_array), 1,          \
                                %(bdescr.tid)d, 103,          \
                                descr=malloc_array_descr)
            check_memory_error(p0)
            jump()
        """)

    def test_rewrite_assembler_maximal_size_2(self):
        self.gc_ll_descr.max_size_of_young_obj = 300
        self.check_rewrite("""
            []
            p0 = new_array(101, descr=bdescr)
            p1 = new_array(102, descr=bdescr)  # two new_arrays can be combined
            p2 = new_array(103, descr=bdescr)  # but not all three
            jump()
        """, """
            []
            p0 = call_malloc_nursery(    \
                              %(2 * (bdescr.basesize + 104))d)
            gc_store(p0, 0,  8765, %(tiddescr.field_size)s)
            gc_store(p0, 0,  101, %(blendescr.field_size)s)
            p1 = nursery_ptr_increment(p0, %(bdescr.basesize + 104)d)
            gc_store(p1, 0,  8765, %(tiddescr.field_size)s)
            gc_store(p1, 0,  102, %(blendescr.field_size)s)
            p2 = call_malloc_nursery(    \
                              %(bdescr.basesize + 104)d)
            gc_store(p2, 0,  8765, %(tiddescr.field_size)s)
            gc_store(p2, 0,  103, %(blendescr.field_size)s)
            jump()
        """)

    def test_rewrite_assembler_huge_size(self):
        # "huge" is defined as "larger than 0xffffff bytes, or 16MB"
        self.check_rewrite("""
            []
            p0 = new_array(20000000, descr=bdescr)
            jump()
        """, """
            []
            p0 = call_r(ConstClass(malloc_array), 1,         \
                                %(bdescr.tid)d, 20000000,    \
                                descr=malloc_array_descr)
            check_memory_error(p0)
            jump()
        """)

    def test_new_with_vtable(self):
        self.check_rewrite("""
            []
            p0 = new_with_vtable(descr=o_descr)
            jump()
        """, """
            [p1]
            p0 = call_malloc_nursery(104)      # rounded up
            gc_store(p0, 0,  9315, %(tiddescr.field_size)s)
            gc_store(p0, 0,  0, %(vtable_descr.field_size)s)
            jump()
        """)

    def test_new_with_vtable_too_big(self):
        self.gc_ll_descr.max_size_of_young_obj = 100
        self.check_rewrite("""
            []
            p0 = new_with_vtable(descr=o_descr)
            jump()
        """, """
            [p1]
            p0 = call_r(ConstClass(malloc_big_fixedsize), 104, 9315, \
                                descr=malloc_big_fixedsize_descr)
            check_memory_error(p0)
            gc_store(p0, 0,  0, %(vtable_descr.field_size)s)
            jump()
        """)

    def test_rewrite_assembler_newstr_newunicode(self):
        # note: strdescr.basesize already contains the extra final character,
        # so that's why newstr(14) is rounded up to 'basesize+15' and not
        # 'basesize+16'.
        self.check_rewrite("""
            [i2]
            p0 = newstr(14)
            p1 = newunicode(10)
            p2 = newunicode(i2)
            p3 = newstr(i2)
            jump()
        """, """
            [i2]
            p0 = call_malloc_nursery(                                \
                      %(strdescr.basesize + 15 * strdescr.itemsize + \
                        unicodedescr.basesize + 10 * unicodedescr.itemsize)d)
            gc_store(p0, 0,  %(strdescr.tid)d, %(tiddescr.field_size)s)
            gc_store(p0, %(strlendescr.offset)s, 14, %(strlendescr.field_size)s)
            gc_store(p0, 0,  0, %(strhashdescr.field_size)s)
            p1 = nursery_ptr_increment(p0, %(strdescr.basesize + 15 * strdescr.itemsize)d)
            gc_store(p1, 0,  %(unicodedescr.tid)d, %(tiddescr.field_size)s)
            gc_store(p1, %(unicodelendescr.offset)s, 10, %(unicodelendescr.field_size)s)
            gc_store(p1, 0,  0, %(unicodehashdescr.field_size)s)
            p2 = call_malloc_nursery_varsize(2, %(unicodedescr.itemsize)d, i2,\
                                descr=unicodedescr)
            gc_store(p2, %(unicodelendescr.offset)s, i2, %(unicodelendescr.field_size)s)
            gc_store(p2, 0,  0, %(unicodehashdescr.field_size)s)
            p3 = call_malloc_nursery_varsize(1, 1, i2, \
                                descr=strdescr)
            gc_store(p3, %(strlendescr.offset)s, i2, %(strlendescr.field_size)s)
            gc_store(p3, 0,  0, %(strhashdescr.field_size)s)
            jump()
        """)

    def test_write_barrier_before_setfield_gc(self):
        self.check_rewrite("""
            [p1, p2]
            setfield_gc(p1, p2, descr=tzdescr)
            jump()
        """, """
            [p1, p2]
            cond_call_gc_wb(p1, descr=wbdescr)
            gc_store(p1, %(tzdescr.offset)s, p2, %(tzdescr.field_size)s)
            jump()
        """)

    def test_write_barrier_before_array_without_from_array(self):
        self.gc_ll_descr.write_barrier_descr.has_write_barrier_from_array = (
            lambda cpu: False)
        self.check_rewrite("""
            [p1, i2, p3]
            setarrayitem_gc(p1, i2, p3, descr=cdescr)
            jump()
        """, """
            [p1, i2, p3]
            cond_call_gc_wb(p1, descr=wbdescr)
            %(setarrayitem('p1', 'i2', 'p3', cdescr))s
            jump()
        """)

    def test_write_barrier_before_short_array(self):
        self.gc_ll_descr.max_size_of_young_obj = 2000
        self.check_rewrite("""
            [i2, p3]
            p1 = new_array_clear(129, descr=cdescr)
            call_n(123456)
            setarrayitem_gc(p1, i2, p3, descr=cdescr)
            jump()
        """, """
            [i2, p3]
            p1 = call_malloc_nursery(    \
                                %(cdescr.basesize + 129 * cdescr.itemsize)d)
            gc_store(p1, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p1, 0,  129, %(clendescr.field_size)s)
            %(zero_array('p1', 0, 129, 'cdescr', cdescr))s
            call_n(123456)
            cond_call_gc_wb(p1, descr=wbdescr)
            %(setarrayitem('p1', 'i2', 'p3', cdescr))s
            jump()
        """)

    def test_write_barrier_before_long_array(self):
        # the limit of "being too long" is fixed, arbitrarily, at 130
        self.gc_ll_descr.max_size_of_young_obj = 2000
        self.check_rewrite("""
            [i2, p3]
            p1 = new_array_clear(130, descr=cdescr)
            call_n(123456)
            setarrayitem_gc(p1, i2, p3, descr=cdescr)
            jump()
        """, """
            [i2, p3]
            p1 = call_malloc_nursery(    \
                                %(cdescr.basesize + 130 * cdescr.itemsize)d)
            gc_store(p1, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p1, 0,  130, %(clendescr.field_size)s)
            %(zero_array('p1', 0, 130, 'cdescr', cdescr))s
            call_n(123456)
            cond_call_gc_wb_array(p1, i2, descr=wbdescr)
            %(setarrayitem('p1', 'i2', 'p3', cdescr))s
            jump()
        """)

    def test_write_barrier_before_unknown_array(self):
        self.check_rewrite("""
            [p1, i2, p3]
            setarrayitem_gc(p1, i2, p3, descr=cdescr)
            jump()
        """, """
            [p1, i2, p3]
            cond_call_gc_wb_array(p1, i2, descr=wbdescr)
            %(setarrayitem('p1', 'i2', 'p3', cdescr))s
            jump()
        """)

    def test_label_makes_size_unknown(self):
        self.check_rewrite("""
            [i2, p3]
            p1 = new_array_clear(5, descr=cdescr)
            label(p1, i2, p3)
            setarrayitem_gc(p1, i2, p3, descr=cdescr)
            jump()
        """, """
            [i2, p3]
            p1 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p1, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p1, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p1', 0, 5, 'cdescr', cdescr))s
            label(p1, i2, p3)
            cond_call_gc_wb_array(p1, i2, descr=wbdescr)
            %(setarrayitem('p1', 'i2', 'p3', cdescr))s
            jump()
        """)

    def test_write_barrier_before_setinteriorfield_gc(self):
        S1 = lltype.GcStruct('S1')
        INTERIOR = lltype.GcArray(('z', lltype.Ptr(S1)))
        interiordescr = get_array_descr(self.gc_ll_descr, INTERIOR)
        interiordescr.tid = 1291
        interiorlendescr = interiordescr.lendescr
        interiorzdescr = get_interiorfield_descr(self.gc_ll_descr,
                                                 INTERIOR, 'z')
        scale = interiorzdescr.arraydescr.itemsize
        offset = interiorzdescr.arraydescr.basesize
        offset += interiorzdescr.fielddescr.offset
        size = interiorzdescr.arraydescr.itemsize
        self.check_rewrite("""
            [p1, p2]
            setinteriorfield_gc(p1, 7, p2, descr=interiorzdescr)
            jump(p1, p2)
        """, """
            [p1, p2]
            cond_call_gc_wb_array(p1, 7, descr=wbdescr)
            gc_store(p1, %(offset + 7 * scale)s, p2, %(size)s)
            jump(p1, p2)
        """, interiorzdescr=interiorzdescr, scale=scale,
             offset=offset, size=size)

    def test_initialization_store(self):
        self.check_rewrite("""
            [p1]
            p0 = new(descr=tdescr)
            setfield_gc(p0, p1, descr=tzdescr)
            jump()
        """, """
            [p1]
            p0 = call_malloc_nursery(%(tdescr.size)d)
            gc_store(p0, 0,  5678, %(tiddescr.field_size)s)
            gc_store(p0, %(tzdescr.offset)s, p1, %(tzdescr.field_size)s)
            jump()
        """)

    def test_initialization_store_2(self):
        self.check_rewrite("""
            []
            p0 = new(descr=tdescr)
            p1 = new(descr=sdescr)
            setfield_gc(p0, p1, descr=tzdescr)
            jump()
        """, """
            []
            p0 = call_malloc_nursery(%(tdescr.size + sdescr.size)d)
            gc_store(p0, 0,  5678, %(tiddescr.field_size)s)
            p1 = nursery_ptr_increment(p0, %(tdescr.size)d)
            gc_store(p1, 0,  1234, %(tiddescr.field_size)s)
            # <<<no cond_call_gc_wb here>>>
            gc_store(p0, %(tzdescr.offset)s, p1, %(tzdescr.field_size)s)
            jump()
        """)

    def test_initialization_store_array(self):
        self.check_rewrite("""
            [p1, i2]
            p0 = new_array_clear(5, descr=cdescr)
            setarrayitem_gc(p0, i2, p1, descr=cdescr)
            jump()
        """, """
            [p1, i2]
            p0 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p0, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p0, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p0', 0, 5, 'cdescr', cdescr))s
            %(setarrayitem('p0', 'i2', 'p1', cdescr))s
            jump()
        """)

    def test_zero_array_reduced_left(self):
        self.check_rewrite("""
            [p1, p2]
            p0 = new_array_clear(5, descr=cdescr)
            setarrayitem_gc(p0, 1, p1, descr=cdescr)
            setarrayitem_gc(p0, 0, p2, descr=cdescr)
            jump()
        """, """
            [p1, p2]
            p0 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p0, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p0, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p0', 2, 3, 'cdescr', cdescr))s
            %(setarrayitem('p0', 1, 'p1', cdescr))s
            %(setarrayitem('p0', 0, 'p2', cdescr))s
            jump()
        """)

    def test_zero_array_reduced_right(self):
        self.check_rewrite("""
            [p1, p2]
            p0 = new_array_clear(5, descr=cdescr)
            setarrayitem_gc(p0, 3, p1, descr=cdescr)
            setarrayitem_gc(p0, 4, p2, descr=cdescr)
            jump()
        """, """
            [p1, p2]
            p0 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p0, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p0, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p0', 0, 3, 'cdescr', cdescr))s
            %(setarrayitem('p0', 3, 'p1', cdescr))s
            %(setarrayitem('p0', 4, 'p2', cdescr))s
            jump()
        """)

    def test_zero_array_not_reduced_at_all(self):
        self.check_rewrite("""
            [p1, p2]
            p0 = new_array_clear(5, descr=cdescr)
            setarrayitem_gc(p0, 3, p1, descr=cdescr)
            setarrayitem_gc(p0, 2, p2, descr=cdescr)
            setarrayitem_gc(p0, 1, p2, descr=cdescr)
            jump()
        """, """
            [p1, p2]
            p0 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p0, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p0, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p0', 0, 5, 'cdescr', cdescr))s
            %(setarrayitem('p0', 3, 'p1', cdescr))s
            %(setarrayitem('p0', 2, 'p2', cdescr))s
            %(setarrayitem('p0', 1, 'p2', cdescr))s
            jump()
        """)

    def test_zero_array_reduced_completely(self):
        self.check_rewrite("""
            [p1, p2]
            p0 = new_array_clear(5, descr=cdescr)
            setarrayitem_gc(p0, 3, p1, descr=cdescr)
            setarrayitem_gc(p0, 4, p2, descr=cdescr)
            setarrayitem_gc(p0, 0, p1, descr=cdescr)
            setarrayitem_gc(p0, 2, p2, descr=cdescr)
            setarrayitem_gc(p0, 1, p2, descr=cdescr)
            jump()
        """, """
            [p1, p2]
            p0 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p0, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p0, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p0', 5, 0, 'cdescr', cdescr))s
            %(setarrayitem('p0', 3, 'p1', cdescr))s
            %(setarrayitem('p0', 4, 'p2', cdescr))s
            %(setarrayitem('p0', 0, 'p1', cdescr))s
            %(setarrayitem('p0', 2, 'p2', cdescr))s
            %(setarrayitem('p0', 1, 'p2', cdescr))s
            jump()
        """)

    def test_zero_array_reduced_left_with_call(self):
        self.check_rewrite("""
            [p1, p2]
            p0 = new_array_clear(5, descr=cdescr)
            setarrayitem_gc(p0, 0, p1, descr=cdescr)
            call_n(321321)
            setarrayitem_gc(p0, 1, p2, descr=cdescr)
            jump()
        """, """
            [p1, p2]
            p0 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p0, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p0, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p0', 1, 4, 'cdescr', cdescr))s
            %(setarrayitem('p0', 0, 'p1', cdescr))s
            call_n(321321)
            cond_call_gc_wb(p0, descr=wbdescr)
            %(setarrayitem('p0', 1, 'p2', cdescr))s
            jump()
        """)

    def test_zero_array_reduced_left_with_label(self):
        self.check_rewrite("""
            [p1, p2]
            p0 = new_array_clear(5, descr=cdescr)
            setarrayitem_gc(p0, 0, p1, descr=cdescr)
            label(p0, p2)
            setarrayitem_gc(p0, 1, p2, descr=cdescr)
            jump()
        """, """
            [p1, p2]
            p0 = call_malloc_nursery(    \
                                %(cdescr.basesize + 5 * cdescr.itemsize)d)
            gc_store(p0, 0,  8111, %(tiddescr.field_size)s)
            gc_store(p0, 0,  5, %(clendescr.field_size)s)
            %(zero_array('p0', 1, 4, 'cdescr', cdescr))s
            %(setarrayitem('p0', 0, 'p1', cdescr))s
            label(p0, p2)
            cond_call_gc_wb_array(p0, 1, descr=wbdescr)
            %(setarrayitem('p0', 1, 'p2', cdescr))s
            jump()
        """)

    def test_zero_array_varsize(self):
        self.check_rewrite("""
            [p1, p2, i3]
            p0 = new_array_clear(i3, descr=bdescr)
            jump()
        """, """
            [p1, p2, i3]
            p0 = call_malloc_nursery_varsize(0, 1, i3, descr=bdescr)
            gc_store(p0, 0,  i3, %(blendescr.field_size)s)
            %(zero_array('p0', 0, 'i3', 'bdescr', bdescr))s
            jump()
        """)

    def test_zero_array_varsize_cannot_reduce(self):
        self.check_rewrite("""
            [p1, p2, i3]
            p0 = new_array_clear(i3, descr=bdescr)
            setarrayitem_gc(p0, 0, p1, descr=bdescr)
            jump()
        """, """
            [p1, p2, i3]
            p0 = call_malloc_nursery_varsize(0, 1, i3, descr=bdescr)
            gc_store(p0, 0,  i3, %(blendescr.field_size)s)
            %(zero_array('p0', 0, 'i3', 'bdescr', bdescr))s
            cond_call_gc_wb_array(p0, 0, descr=wbdescr)
            %(setarrayitem('p0', 0, 'p1', bdescr))s
            jump()
        """)

    def test_initialization_store_potentially_large_array(self):
        # the write barrier cannot be omitted, because we might get
        # an array with cards and the GC assumes that the write
        # barrier is always called, even on young (but large) arrays
        self.check_rewrite("""
            [i0, p1, i2]
            p0 = new_array(i0, descr=bdescr)
            setarrayitem_gc(p0, i2, p1, descr=bdescr)
            jump()
        """, """
            [i0, p1, i2]
            p0 = call_malloc_nursery_varsize(0, 1, i0, descr=bdescr)
            gc_store(p0, 0,  i0, %(blendescr.field_size)s)
            cond_call_gc_wb_array(p0, i2, descr=wbdescr)
            gc_store_indexed(p0, i2, p1, 1, %(bdescr.basesize)s, 1)
            jump()
        """)

    def test_non_initialization_store(self):
        self.check_rewrite("""
            [i0]
            p0 = new(descr=tdescr)
            p1 = newstr(i0)
            setfield_gc(p0, p1, descr=tzdescr)
            jump()
        """, """
            [i0]
            p0 = call_malloc_nursery(%(tdescr.size)d)
            gc_store(p0, 0,  5678, %(tiddescr.field_size)s)
            gc_store(p0, %(tdescr.gc_fielddescrs[0].offset)s, 0, %(tdescr.gc_fielddescrs[0].offset)s)
            p1 = call_malloc_nursery_varsize(1, 1, i0, \
                                descr=strdescr)
            gc_store(p1, %(strlendescr.offset)s, i0, %(strlendescr.field_size)s)
            gc_store(p1, 0,  0, %(strhashdescr.field_size)s)
            cond_call_gc_wb(p0, descr=wbdescr)
            gc_store(p0, %(tzdescr.offset)s, p1, %(tzdescr.field_size)s)
            jump()
        """)

    def test_non_initialization_store_label(self):
        self.check_rewrite("""
            [p1]
            p0 = new(descr=tdescr)
            label(p0, p1)
            setfield_gc(p0, p1, descr=tzdescr)
            jump()
        """, """
            [p1]
            p0 = call_malloc_nursery(%(tdescr.size)d)
            gc_store(p0, 0,  5678, %(tiddescr.field_size)s)
            gc_store(p0, %(tdescr.gc_fielddescrs[0].offset)s, 0, %(tdescr.gc_fielddescrs[0].offset)s)
            label(p0, p1)
            cond_call_gc_wb(p0, descr=wbdescr)
            gc_store(p0, %(tzdescr.offset)s, p1, %(tzdescr.field_size)s)
            jump()
        """)

    def test_multiple_writes(self):
        self.check_rewrite("""
            [p0, p1, p2]
            setfield_gc(p0, p1, descr=tzdescr)
            setfield_gc(p0, p2, descr=tzdescr)
            jump(p1, p2, p0)
        """, """
            [p0, p1, p2]
            cond_call_gc_wb(p0, descr=wbdescr)
            gc_store(p0, %(tzdescr.offset)s, p1, %(tzdescr.field_size)s)
            gc_store(p0, %(tzdescr.offset)s, p2, %(tzdescr.field_size)s)
            jump(p1, p2, p0)
        """)

    def test_rewrite_call_assembler(self):
        self.check_rewrite("""
        [i0, f0]
        i2 = call_assembler_i(i0, f0, descr=casmdescr)
        """, """
        [i0, f0]
        i1 = gc_load_i(ConstClass(frame_info), %(jfi_frame_size.offset)s, %(jfi_frame_size.field_size)s)
        p1 = call_malloc_nursery_varsize_frame(i1)
        gc_store(p1, 0,  0, %(tiddescr.field_size)s)
        i2 = gc_load_i(ConstClass(frame_info), %(jfi_frame_depth.offset)s, %(jfi_frame_depth.field_size)s)
        %(setfield('p1', 'NULL', jf_savedata))s
        %(setfield('p1', 'NULL', jf_force_descr))s
        %(setfield('p1', 'NULL', jf_descr))s
        %(setfield('p1', 'NULL', jf_guard_exc))s
        %(setfield('p1', 'NULL', jf_forward))s
        gc_store(p1, 0, i2, %(framelendescr.field_size)s)
        %(setfield('p1', 'ConstClass(frame_info)', jf_frame_info))s
        gc_store(p1, 3, i0, 8)
        gc_store(p1, 13, f0, 8)
        i3 = call_assembler_i(p1, descr=casmdescr)
        """)

    def test_int_add_ovf(self):
        self.check_rewrite("""
            [i0]
            p0 = new(descr=tdescr)
            i1 = int_add_ovf(i0, 123)
            guard_overflow(descr=guarddescr) []
            jump()
        """, """
            [i0]
            p0 = call_malloc_nursery(%(tdescr.size)d)
            gc_store(p0, 0,  5678, %(tiddescr.field_size)s)
            gc_store(p0, %(tdescr.gc_fielddescrs[0].offset)s, 0, %(tdescr.gc_fielddescrs[0].offset)s)
            i1 = int_add_ovf(i0, 123)
            guard_overflow(descr=guarddescr) []
            jump()
        """)

    def test_int_gt(self):
        self.check_rewrite("""
            [i0]
            p0 = new(descr=tdescr)
            i1 = int_gt(i0, 123)
            guard_false(i1, descr=guarddescr) []
            jump()
        """, """
            [i0]
            p0 = call_malloc_nursery(%(tdescr.size)d)
            gc_store(p0, 0,  5678, %(tiddescr.field_size)s)
            gc_store(p0, %(tdescr.gc_fielddescrs[0].offset)s, 0, %(tdescr.gc_fielddescrs[0].offset)s)
            i1 = int_gt(i0, 123)
            guard_false(i1, descr=guarddescr) []
            jump()
        """)

    def test_zero_ptr_field_before_getfield(self):
        # This case may need to be fixed in the metainterp/optimizeopt
        # already so that it no longer occurs for rewrite.py.  But anyway
        # it's a good idea to make sure rewrite.py is correct on its own.
        self.check_rewrite("""
            []
            p0 = new(descr=tdescr)
            p1 = getfield_gc_r(p0, descr=tzdescr)
            jump(p1)
        """, """
            []
            p0 = call_malloc_nursery(%(tdescr.size)d)
            gc_store(p0, 0,  5678, %(tiddescr.field_size)s)
            gc_store(p0, %(tdescr.gc_fielddescrs[0].offset)s, 0, %(tdescr.gc_fielddescrs[0].offset)s)
            p1 = gc_load_r(p0, %(tzdescr.offset)s, %(tzdescr.field_size)s)
            jump(p1)
        """)

    def test_remove_tested_failarg(self):
        self.check_rewrite("""
            [i5]
            i2 = int_ge(i5, 10)
            guard_true(i2) [i5, i2]
            jump()
        """, """
            [i5]
            i0 = same_as_i(0)
            i2 = int_ge(i5, 10)
            guard_true(i2) [i5, i0]
            jump()
        """)
        self.check_rewrite("""
            [i5]
            i2 = int_ge(i5, 10)
            guard_false(i2) [i5, i2]
            jump()
        """, """
            [i5]
            i0 = same_as_i(1)
            i2 = int_ge(i5, 10)
            guard_false(i2) [i5, i0]
            jump()
        """)

    @py.test.mark.parametrize('support_offset,factors,fromto',[
        # [False, (1,2,4,8), 'setarrayitem_gc(p0,i1,i2,descr=adescr)' '->'
        #    'i3 = int_mul(i1,%(adescr.itemsize)s);'
        #    'i4 = int_add(i3,%(adescr.basesize)s);'
        #    'gc_store(p0,i4,i2,%(adescr.itemsize)s)'],
        [True, (1,2,4,8), 'setarrayitem_gc(p0,i1,i2,descr=adescr)' '->'
           'gc_store_indexed(p0,i1,i2,%(adescr.itemsize)s,'
           '%(adescr.basesize)s,%(adescr.itemsize)s)'],
        #[False, (1,), 'setarrayitem_gc(p0,i1,i2,descr=adescr)' '->'
        #   'i3 = int_mul(i1,%(adescr.itemsize)s);'
        #   'i4 = int_add(i3,%(adescr.basesize)s);'
        #   'gc_store(p0,i4,i2,%(adescr.itemsize)s)'],
        [True, None, 'i3 = raw_load_i(p0,i1,descr=adescr)' '->'
           'gc_load_indexed_i(p0,i1,1,%(adescr.basesize)s,-%(adescr.itemsize)s)'],
        [True, None, 'i3 = raw_load_f(p0,i1,descr=fdescr)' '->'
           'gc_load_indexed_f(p0,i1,1,%(fdescr.basesize)s,%(fdescr.itemsize)s)'],
        [True, None, 'i3 = raw_load_i(p0,i1,descr=sfdescr)' '->'
           'gc_load_indexed_i(p0,i1,1,%(sfdescr.basesize)s,%(sfdescr.itemsize)s)'],
        [True, (1,2,4,8), 'i3 = raw_store(p0,i1,i2,descr=raw_sfdescr)' '->'
           'gc_store_indexed(p0,i1,i2,1,%(raw_sfdescr.basesize)s,%(raw_sfdescr.itemsize)s)'],
        # [False, (1,), 'i3 = raw_store(p0,i1,i2,descr=raw_sfdescr)' '->'
        #    'i5 = int_add(i1,%(raw_sfdescr.basesize)s);'
        #    'gc_store(p0,i5,i2,%(raw_sfdescr.itemsize)s)'],
        [True, (1,2,4,8), 'i3 = getfield_gc_f(p0,descr=ydescr)' '->'
           'i3 = gc_load_f(p0,%(ydescr.offset)s,%(ydescr.field_size)s)'],
        [True, (1,2,4,8), 'setfield_raw(p0,i1,descr=ydescr)' '->'
           'gc_store(p0,%(ydescr.offset)s,i1,%(ydescr.field_size)s)'],
        [True, (1,2,4,8), 'setfield_gc(p0,p0,descr=zdescr)' '->'
           'cond_call_gc_wb(p0, descr=wbdescr);'
           'gc_store(p0,%(zdescr.offset)s,p0,%(zdescr.field_size)s)'],
        [False, (1,), 'i3 = arraylen_gc(p0, descr=adescr)' '->'
                      'i3 = gc_load_i(p0,0,%(adescr.itemsize)s)'],
        #[False, (1,),  'i3 = strlen(p0)' '->'
        #               'i3 = gc_load_i(p0,'
        #               '%(strlendescr.offset)s,%(strlendescr.field_size)s)'],
        [True,  (1,),  'i3 = strlen(p0)' '->'
                       'i3 = gc_load_i(p0,'
                                 '%(strlendescr.offset)s,'
                                 '%(strlendescr.field_size)s)'],
        [True,  (1,),  'i3 = strhash(p0)' '->'
                       'i3 = gc_load_i(p0,'
                                 '%(strhashdescr.offset)s,'
                                 '-%(strhashdescr.field_size)s)'],
        #[False, (1,),  'i3 = unicodelen(p0)' '->'
        #               'i3 = gc_load_i(p0,'
        #                       '%(unicodelendescr.offset)s,'
        #                       '%(unicodelendescr.field_size)s)'],
        [True,  (1,),  'i3 = unicodelen(p0)' '->'
                       'i3 = gc_load_i(p0,'
                               '%(unicodelendescr.offset)s,'
                               '%(unicodelendescr.field_size)s)'],
        [True,  (1,),  'i3 = unicodehash(p0)' '->'
                       'i3 = gc_load_i(p0,'
                                 '%(unicodehashdescr.offset)s,'
                                 '-%(unicodehashdescr.field_size)s)'],
        ## getitem str/unicode
        [True,  (2,4), 'i3 = unicodegetitem(p0,i1)' '->'
                       'i3 = gc_load_indexed_i(p0,i1,'
                                  '%(unicodedescr.itemsize)d,'
                                  '%(unicodedescr.basesize)d,'
                                  '%(unicodedescr.itemsize)d)'],
        #[False, (2,4), 'i3 = unicodegetitem(p0,i1)' '->'
        #               'i4 = int_mul(i1, %(unicodedescr.itemsize)d);'
        #               'i5 = int_add(i4, %(unicodedescr.basesize)d);'
        #               'i3 = gc_load_i(p0,i5,%(unicodedescr.itemsize)d)'],
        [True,  (4,),  'i3 = strgetitem(p0,i1)' '->'
                       'i3 = gc_load_indexed_i(p0,i1,1,'
                       '%(strdescr.basesize-1)d,1)'],
        #[False, (4,),  'i3 = strgetitem(p0,i1)' '->'
        #               'i5 = int_add(i1, %(strdescr.basesize-1)d);'
        #               'i3 = gc_load_i(p0,i5,1)'],
        ## setitem str/unicode
        [True, (4,),  'i3 = strsetitem(p0,i1,0)' '->'
                      'i3 = gc_store_indexed(p0,i1,0,1,'
                               '%(strdescr.basesize-1)d,1)'],
        [True, (2,4), 'i3 = unicodesetitem(p0,i1,0)' '->'
                      'i3 = gc_store_indexed(p0,i1,0,'
                                 '%(unicodedescr.itemsize)d,'
                                 '%(unicodedescr.basesize)d,'
                                 '%(unicodedescr.itemsize)d)'],
        ## interior
        [True, (1,2,4,8), 'i3 = getinteriorfield_gc_i(p0,i1,descr=itzdescr)' '->'
                          'i4 = int_mul(i1,'
                             '%(itzdescr.arraydescr.itemsize)d);'
                          'i3 = gc_load_indexed_i(p0,i4,1,'
                                   '%(itzdescr.arraydescr.basesize'
                                   '   + itzdescr.fielddescr.offset)d,'
                                   '%(itzdescr.fielddescr.field_size)d)'],
        [True, (1,2,4,8), 'i3 = getinteriorfield_gc_r(p0,i1,descr=itxdescr)' '->'
                          'i4 = int_mul(i1,'
                             '%(itxdescr.arraydescr.itemsize)d);'
                          'i3 = gc_load_indexed_r(p0,i4,1,'
                             '%(itxdescr.arraydescr.basesize'
                             '   + itxdescr.fielddescr.offset)d,'
                             '%(itxdescr.fielddescr.field_size)d)'],
        [True, (1,2,4,8), 'i3 = setinteriorfield_gc(p0,i1,i2,descr=itydescr)' '->'
                          'i4 = int_mul(i1,'
                             '%(itydescr.arraydescr.itemsize)d);'
                          'i3 = gc_store_indexed(p0,i4,i2,1,'
                             '%(itydescr.arraydescr.basesize'
                             '   + itydescr.fielddescr.offset)d,'
                             '%(itydescr.fielddescr.field_size)d)'],
        [True, (1,2,4,8), 'i3 = setinteriorfield_gc(p0,i1,i2,descr=ity2descr)' '->'
                          'i4 = int_lshift(i1,'
                             '%(s2i_item_size_in_bits)d);'
                          'i3 = gc_store_indexed(p0,i4,i2,1,'
                             '%(ity2descr.arraydescr.basesize'
                             '   + itydescr.fielddescr.offset)d,'
                             '%(ity2descr.fielddescr.field_size)d)'],
        [True, (2,), 'i3 = gc_load_indexed_i(i1, i2, 2, 40, 1)' '->'
                     'i3 = gc_load_indexed_i(i1, i2, 2, 40, 1)'],
        [True, (2,), 'i3 = gc_load_indexed_i(i1, 6, 8, 40, 4)' '->'
                     'i3 = gc_load_i(i1, 88, 4)'],
        [True, (2,), 'i3 = gc_load_indexed_i(i1, i2, 2, 40, -1)' '->'
                     'i3 = gc_load_indexed_i(i1, i2, 2, 40, -1)'],
        [True, (2,), 'i3 = gc_load_indexed_i(i1, 6, 8, 40, -4)' '->'
                     'i3 = gc_load_i(i1, 88, -4)'],
        [True, (2,), 'i3 = gc_store_indexed(i1, i2, 999, 2, 40, 1)' '->'
                     'i3 = gc_store_indexed(i1, i2, 999, 2, 40, 1)'],
        [True, (2,), 'i3 = gc_store_indexed(i1, 6, 999, 8, 40, 2)' '->'
                     'i3 = gc_store(i1, 88, 999, 2)'],
    ])
    def test_gc_load_store_transform(self, support_offset, factors, fromto):
        self.cpu.load_constant_offset = support_offset
        all_supported_sizes = [factors]

        if not factors:
            all_supported_sizes = [(1,), (1,2,), (4,), (1,2,4,8)]
        try:
            for factors in all_supported_sizes:
                self.cpu.load_supported_factors = factors
                f, t = fromto.split('->')
                t = ('\n' +(' '*20)).join([s for s in t.split(';')])
                self.check_rewrite("""
                    [p0,i1,i2]
                    {f}
                    jump()
                """.format(**locals()), """
                    [p0,i1,i2]
                    {t}
                    jump()
                """.format(**locals()))
        finally:
            del self.cpu.load_supported_factors   # restore class-level value

    def test_load_from_gc_table_1i(self):
        self.check_rewrite("""
            [i1]
            setfield_gc(ConstPtr(myR1), i1, descr=xdescr)
            jump()
        """, """
            [i1]
            p0 = load_from_gc_table(0)
            gc_store(p0, %(xdescr.offset)s, i1, %(xdescr.field_size)s)
            jump()
        """)
        assert self.gcrefs == [self.myR1]

    def test_load_from_gc_table_1p(self):
        self.check_rewrite("""
            [p1]
            setfield_gc(ConstPtr(myT), p1, descr=tzdescr)
            jump()
        """, """
            [i1]
            p0 = load_from_gc_table(0)
            cond_call_gc_wb(p0, descr=wbdescr)
            gc_store(p0, %(tzdescr.offset)s, i1, %(tzdescr.field_size)s)
            jump()
        """)
        assert self.gcrefs == [self.myT]

    def test_load_from_gc_table_2(self):
        self.check_rewrite("""
            [i1, f2]
            setfield_gc(ConstPtr(myR1), i1, descr=xdescr)
            setfield_gc(ConstPtr(myR1), f2, descr=ydescr)
            jump()
        """, """
            [i1, f2]
            p0 = load_from_gc_table(0)
            gc_store(p0, %(xdescr.offset)s, i1, %(xdescr.field_size)s)
            gc_store(p0, %(ydescr.offset)s, f2, %(ydescr.field_size)s)
            jump()
        """)
        assert self.gcrefs == [self.myR1]

    def test_load_from_gc_table_3(self):
        self.check_rewrite("""
            [i1, f2]
            setfield_gc(ConstPtr(myR1), i1, descr=xdescr)
            label(f2)
            setfield_gc(ConstPtr(myR1), f2, descr=ydescr)
            jump()
        """, """
            [i1, f2]
            p0 = load_from_gc_table(0)
            gc_store(p0, %(xdescr.offset)s, i1, %(xdescr.field_size)s)
            label(f2)
            p1 = load_from_gc_table(0)
            gc_store(p1, %(ydescr.offset)s, f2, %(ydescr.field_size)s)
            jump()
        """)
        assert self.gcrefs == [self.myR1]

    def test_load_from_gc_table_4(self):
        self.check_rewrite("""
            [i1, f2]
            setfield_gc(ConstPtr(myR1), i1, descr=xdescr)
            setfield_gc(ConstPtr(myR1b), f2, descr=ydescr)
            jump()
        """, """
            [i1, f2]
            p0 = load_from_gc_table(0)
            gc_store(p0, %(xdescr.offset)s, i1, %(xdescr.field_size)s)
            p1 = load_from_gc_table(1)
            gc_store(p1, %(ydescr.offset)s, f2, %(ydescr.field_size)s)
            jump()
        """)
        assert self.gcrefs == [self.myR1, self.myR1b]

    def test_pinned_simple_getfield(self):
        # originally in test_pinned_object_rewrite; now should give the
        # same result for pinned objects and for normal objects
        self.check_rewrite("""
            []
            i0 = getfield_gc_i(ConstPtr(myR1), descr=xdescr)
        """, """
            []
            p1 = load_from_gc_table(0)
            i0 = gc_load_i(p1, %(xdescr.offset)s, -%(xdescr.field_size)s)
        """)
        assert self.gcrefs == [self.myR1]

    def test_pinned_simple_getfield_twice(self):
        # originally in test_pinned_object_rewrite; now should give the
        # same result for pinned objects and for normal objects
        self.check_rewrite("""
            []
            i0 = getfield_gc_i(ConstPtr(myR1), descr=xdescr)
            i1 = getfield_gc_i(ConstPtr(myR1b), descr=xdescr)
            i2 = getfield_gc_i(ConstPtr(myR1), descr=xdescr)
        """, """
            []
            p1 = load_from_gc_table(0)
            i0 = gc_load_i(p1, %(xdescr.offset)s, -%(xdescr.field_size)s)
            p2 = load_from_gc_table(1)
            i1 = gc_load_i(p2, %(xdescr.offset)s, -%(xdescr.field_size)s)
            i2 = gc_load_i(p1, %(xdescr.offset)s, -%(xdescr.field_size)s)
        """)
        assert self.gcrefs == [self.myR1, self.myR1b]

    def test_guard_in_gcref(self):
        self.check_rewrite("""
            [i1, i2]
            guard_true(i1) []
            guard_true(i2) []
            jump()
        """, """
            [i1, i2]
            guard_true(i1) []
            guard_true(i2) []
            jump()
        """)
        assert len(self.gcrefs) == 2

    def test_rewrite_copystrcontents(self):
        self.check_rewrite("""
        [p0, p1, i0, i1, i_len]
        copystrcontent(p0, p1, i0, i1, i_len)
        """, """
        [p0, p1, i0, i1, i_len]
        i2 = load_effective_address(p0, i0, %(str_basesize)s, 0)
        i3 = load_effective_address(p1, i1, %(str_basesize)s, 0)
        call_n(ConstClass(memcpy_fn), i3, i2, i_len, descr=memcpy_descr)
        """)

    def test_rewrite_copystrcontents_without_load_effective_address(self):
        self.cpu.supports_load_effective_address = False
        self.check_rewrite("""
        [p0, p1, i0, i1, i_len]
        copystrcontent(p0, p1, i0, i1, i_len)
        """, """
        [p0, p1, i0, i1, i_len]
        i2b = int_add(p0, i0)
        i2 = int_add(i2b, %(str_basesize)s)
        i3b = int_add(p1, i1)
        i3 = int_add(i3b, %(str_basesize)s)
        call_n(ConstClass(memcpy_fn), i3, i2, i_len, descr=memcpy_descr)
        """)

    def test_rewrite_copyunicodecontents(self):
        self.check_rewrite("""
        [p0, p1, i0, i1, i_len]
        copyunicodecontent(p0, p1, i0, i1, i_len)
        """, """
        [p0, p1, i0, i1, i_len]
        i2 = load_effective_address(p0, i0, %(uni_basesize)s, %(uni_itemscale)d)
        i3 = load_effective_address(p1, i1, %(uni_basesize)s, %(uni_itemscale)d)
        i4 = int_lshift(i_len, %(uni_itemscale)d)
        call_n(ConstClass(memcpy_fn), i3, i2, i4, descr=memcpy_descr)
        """)

    def test_rewrite_copyunicodecontents_without_load_effective_address(self):
        self.cpu.supports_load_effective_address = False
        self.check_rewrite("""
        [p0, p1, i0, i1, i_len]
        copyunicodecontent(p0, p1, i0, i1, i_len)
        """, """
        [p0, p1, i0, i1, i_len]
        i0s = int_lshift(i0, %(uni_itemscale)d)
        i2b = int_add(p0, i0s)
        i2 = int_add(i2b, %(uni_basesize)s)
        i1s = int_lshift(i1, %(uni_itemscale)d)
        i3b = int_add(p1, i1s)
        i3 = int_add(i3b, %(uni_basesize)s)
        i4 = int_lshift(i_len, %(uni_itemscale)d)
        call_n(ConstClass(memcpy_fn), i3, i2, i4, descr=memcpy_descr)
        """)

    def test_record_int_add_or_sub(self):
        # ---- no rewrite ----
        self.check_rewrite("""
            [p0, i0]
            i2 = getarrayitem_gc_i(p0, i0, descr=cdescr)
        """, """
            [p0, i0]
            i3 = gc_load_indexed_i(p0, i0,   \
                      %(cdescr.itemsize)d,   \
                      %(cdescr.basesize)d,   \
                      %(cdescr.itemsize)d)
        """)
        # ---- add 5 ----
        self.check_rewrite("""
            [p0, i0]
            i1 = int_add(i0, 5)
            i2 = getarrayitem_gc_i(p0, i1, descr=cdescr)
        """, """
            [p0, i0]
            i1 = int_add(i0, 5)
            i3 = gc_load_indexed_i(p0, i0,   \
                      %(cdescr.itemsize)d,   \
                      %(cdescr.basesize + 5 * cdescr.itemsize)d,   \
                      %(cdescr.itemsize)d)
        """)
        # ---- subtract 1 ----
        self.check_rewrite("""
            [p0, i0]
            i1 = int_sub(i0, 1)
            i2 = getarrayitem_gc_i(p0, i1, descr=cdescr)
        """, """
            [p0, i0]
            i1 = int_sub(i0, 1)
            i3 = gc_load_indexed_i(p0, i0,   \
                      %(cdescr.itemsize)d,   \
                      %(cdescr.basesize - cdescr.itemsize)d,   \
                      %(cdescr.itemsize)d)
        """)
        # ---- add reversed and multiple levels ----
        self.check_rewrite("""
            [p0, i0]
            i1 = int_sub(i0, 1)
            i2 = int_add(i1, 10)
            i3 = int_add(100, i2)
            i4 = getarrayitem_gc_i(p0, i3, descr=cdescr)
        """, """
            [p0, i0]
            i1 = int_sub(i0, 1)
            i2 = int_add(i1, 10)
            i3 = int_add(100, i2)
            i4 = gc_load_indexed_i(p0, i0,   \
                      %(cdescr.itemsize)d,   \
                      %(cdescr.basesize + 109 * cdescr.itemsize)d,   \
                      %(cdescr.itemsize)d)
        """)
        # ---- a label stops the optimization ----
        self.check_rewrite("""
            [p0, i0]
            i1 = int_sub(i0, 1)
            label(p0, i0, i1)
            i4 = getarrayitem_gc_i(p0, i1, descr=cdescr)
        """, """
            [p0, i0]
            i1 = int_sub(i0, 1)
            label(p0, i0, i1)
            i4 = gc_load_indexed_i(p0, i1,   \
                      %(cdescr.itemsize)d,   \
                      %(cdescr.basesize)d,   \
                      %(cdescr.itemsize)d)
        """)
        # ---- also test setarrayitem_gc ----
        self.check_rewrite("""
            [p0, i0, i4]
            i1 = int_sub(i0, 1)
            i2 = int_add(i1, 10)
            i3 = int_add(100, i2)
            setarrayitem_gc(p0, i3, i4, descr=cdescr)
        """, """
            [p0, i0, i4]
            i1 = int_sub(i0, 1)
            i2 = int_add(i1, 10)
            i3 = int_add(100, i2)
            gc_store_indexed(p0, i0, i4,   \
                      %(cdescr.itemsize)d,   \
                      %(cdescr.basesize + 109 * cdescr.itemsize)d,   \
                      %(cdescr.itemsize)d)
        """)
        # ---- also check int_add_ovf, int_sub_ovf ----
        self.check_rewrite("""
            [p0, i0, i4]
            i1 = int_sub_ovf(i0, 1)
            guard_no_overflow(descr=guarddescr) []
            i2 = int_add_ovf(i1, 10)
            guard_no_overflow(descr=guarddescr) []
            i3 = int_add_ovf(100, i2)
            guard_no_overflow(descr=guarddescr) []
            setarrayitem_gc(p0, i3, i4, descr=cdescr)
        """, """
            [p0, i0, i4]
            i1 = int_sub_ovf(i0, 1)
            guard_no_overflow(descr=guarddescr) []
            i2 = int_add_ovf(i1, 10)
            guard_no_overflow(descr=guarddescr) []
            i3 = int_add_ovf(100, i2)
            guard_no_overflow(descr=guarddescr) []
            gc_store_indexed(p0, i0, i4,   \
                      %(cdescr.itemsize)d,   \
                      %(cdescr.basesize + 109 * cdescr.itemsize)d,   \
                      %(cdescr.itemsize)d)
        """)

    def test_guard_always_fails(self):
        self.check_rewrite("""
        [i1, i2, i3]
        guard_always_fails(descr=guarddescr) [i1, i2, i3]
        """, """
        [i1, i2, i3]
        i4 = same_as_i(0)
        guard_value(i4, 1, descr=guarddescr) [i1, i2, i3]
        """)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import py
import sys, os, signal, thread, time, codecs
from dotviewer.conftest import option
from dotviewer.strunicode import RAW_ENCODING

SOURCE1 = u"""digraph G{
λ -> b [label="λλλλλ"]
b -> μ
}
"""

FILENAME = 'graph1.dot'

def setup_module(mod):
    if not option.pygame:
        py.test.skip("--pygame not enabled")
    udir = py.path.local.make_numbered_dir(prefix='usession-dot-', keep=3)
    f = codecs.open(str(udir.join(FILENAME)), 'wb', RAW_ENCODING)
    f.write(SOURCE1)
    f.close()

    from dotviewer import graphclient
    mod.pkgdir = py.path.local(graphclient.this_dir)
    mod.udir = udir

    try:
        del os.environ['GRAPHSERVER']
    except KeyError:
        pass


def test_dotviewer():
    print "=== dotviewer.py %s" % FILENAME
    err = os.system('"%s" "%s"' % (pkgdir.join('dotviewer.py'),
                                   udir.join(FILENAME)))
    assert err == 0

    plain_name = FILENAME.replace('.dot','.plain')
    
    os.system('dot -Tplain "%s" > "%s"' % (udir.join(FILENAME),
                                           udir.join(plain_name)))
    print "=== dotviewer.py %s" % plain_name
    err = os.system('"%s" "%s"' % (pkgdir.join('dotviewer.py'),
                                   udir.join(plain_name)))
    assert err == 0

def test_display_dot_file():
    from dotviewer.graphclient import display_dot_file
    print "=== display_dot_file(%s) with GRAPHSERVER=%s" % (
        FILENAME, os.environ.get('GRAPHSERVER', ''),)
    display_dot_file(udir.join(FILENAME))
    print "=== display_dot_file finished"


def test_graphserver():
    import socket
    s = socket.socket()
    s.listen(1)
    host, port = s.getsockname()     # pick a random free port
    s.close()

    if hasattr(sys, 'pypy_objspaceclass'):
        python = 'python'
    else:
        python = sys.executable

    cmdargs = [python, str(pkgdir.join('graphserver.py')),
               str(port)]
    print '* starting:', ' '.join(cmdargs)
    pid = os.spawnv(os.P_NOWAIT, cmdargs[0], cmdargs)
    try:
        time.sleep(1)    # hack - wait a bit to make sure the server is up
        os.environ['GRAPHSERVER'] = '%s:%d' % (host, port)
        try:
            test_display_dot_file()
        finally:
            del os.environ['GRAPHSERVER']
    finally:
        os.kill(pid, signal.SIGTERM)

def test_colors():
    from dotviewer import graphpage, graphclient
    class MyPage(graphpage.DotFileGraphPage):
        def compute(self, dotfile):
            super(MyPage, self).compute(dotfile)
            self.links = {'v2721': 'Hello world',
                          'v2720': ('Something green', (0, 192, 0)),
                          }
    dotfile = str(udir.join(FILENAME))
    page = MyPage(dotfile)
    graphclient.display_page(page)

def test_fixedfont():
    from dotviewer import graphpage, graphclient
    class MyPage(graphpage.DotFileGraphPage):
        fixedfont = True
    dotfile = str(udir.join(FILENAME))
    page = MyPage(dotfile)
    page.fixedfont = True
    graphclient.display_page(page)



from pypy.module.cpyext.test.test_cpyext import AppTestCpythonExtensionBase
import pytest

class AppTestBufferProtocol(AppTestCpythonExtensionBase):
    """Tests for the old buffer protocol."""

    def w_get_buffer_support(self):
        return self.import_extension('buffer_support', [
            ("charbuffer_as_string", "METH_O",
             """
                 char *ptr;
                 Py_ssize_t size;
                 if (PyObject_AsCharBuffer(args, (const char **)&ptr, &size) < 0)
                     return NULL;
                 return PyString_FromStringAndSize(ptr, size);
             """),
            ("check_readbuffer", "METH_O",
             """
                 return PyBool_FromLong(PyObject_CheckReadBuffer(args));
             """),
            ("readbuffer_as_string", "METH_O",
             """
                 const void *ptr;
                 Py_ssize_t size;
                 if (PyObject_AsReadBuffer(args, &ptr, &size) < 0)
                     return NULL;
                 return PyString_FromStringAndSize((char*)ptr, size);
             """),
            ("writebuffer_as_string", "METH_O",
             """
                 void *ptr;
                 Py_ssize_t size;
                 if (PyObject_AsWriteBuffer(args, &ptr, &size) < 0)
                     return NULL;
                 return PyString_FromStringAndSize((char*)ptr, size);
             """),
            ("zero_out_writebuffer", "METH_O",
             """
                 void *ptr;
                 Py_ssize_t size;
                 Py_ssize_t i;
                 if (PyObject_AsWriteBuffer(args, &ptr, &size) < 0)
                     return NULL;
                 for (i = 0; i < size; i++) {
                     ((char*)ptr)[i] = 0;
                 }
                 Py_RETURN_NONE;
             """),
            ])

    def test_string(self):
        buffer_support = self.get_buffer_support()

        s = 'a\0x'

        assert buffer_support.check_readbuffer(s)
        assert s == buffer_support.readbuffer_as_string(s)
        assert raises(TypeError, buffer_support.writebuffer_as_string, s)
        assert s == buffer_support.charbuffer_as_string(s)

    def test_buffer(self):
        buffer_support = self.get_buffer_support()

        s = 'a\0x'
        buf = buffer(s)

        assert buffer_support.check_readbuffer(buf)
        assert s == buffer_support.readbuffer_as_string(buf)
        assert raises(TypeError, buffer_support.writebuffer_as_string, buf)
        assert s == buffer_support.charbuffer_as_string(buf)

    def test_mmap(self):
        import mmap
        buffer_support = self.get_buffer_support()

        s = 'a\0x'
        mm = mmap.mmap(-1, 3)
        mm[:] = s

        assert buffer_support.check_readbuffer(mm)
        assert s == buffer_support.readbuffer_as_string(mm)
        assert s == buffer_support.writebuffer_as_string(mm)
        assert s == buffer_support.charbuffer_as_string(mm)

        s = '\0' * 3
        buffer_support.zero_out_writebuffer(mm)
        assert s == ''.join(mm)
        assert s == buffer_support.readbuffer_as_string(mm)
        assert s == buffer_support.writebuffer_as_string(mm)
        assert s == buffer_support.charbuffer_as_string(mm)

        s = '\0' * 3
        ro_mm = mmap.mmap(-1, 3, access=mmap.ACCESS_READ)
        assert buffer_support.check_readbuffer(ro_mm)
        assert s == buffer_support.readbuffer_as_string(ro_mm)
        assert raises(TypeError, buffer_support.writebuffer_as_string, ro_mm)
        assert s == buffer_support.charbuffer_as_string(ro_mm)

    def test_array(self):
        import array
        buffer_support = self.get_buffer_support()

        s = 'a\0x'
        a = array.array('B', [5, 0, 10])

        buffer_support.zero_out_writebuffer(a)
        assert list(a) == [0, 0, 0]

    def test_nonbuffer(self):
        # e.g. int
        buffer_support = self.get_buffer_support()

        assert not buffer_support.check_readbuffer(42)
        assert raises(TypeError, buffer_support.readbuffer_as_string, 42)
        assert raises(TypeError, buffer_support.writebuffer_as_string, 42)
        assert raises(TypeError, buffer_support.charbuffer_as_string, 42)

    def test_user_class(self):
        class MyBuf(str):
            pass
        s = 'a\0x'
        buf = MyBuf(s)
        buffer_support = self.get_buffer_support()

        assert buffer_support.check_readbuffer(buf)
        assert s == buffer_support.readbuffer_as_string(buf)
        assert raises(TypeError, buffer_support.writebuffer_as_string, buf)
        assert s == buffer_support.charbuffer_as_string(buf)

    def test_user_bufferable(self):
        try:
            import __pypy__
        except ImportError:
            skip('PyPy only test')

        class MyBuf(__pypy__.bufferable.bufferable):
            def __init__(self, mview):
                if not isinstance(mview, memoryview):
                    raise ValueError('mview must be a memoryview')
                self.mview = mview

            def __buffer__(self, flags):
                if flags & 1 and self.buffer.readonly:
                    raise TypeError('cannot return a writable buffer')
                return __pypy__.newmemoryview(self.mview, self.mview.itemsize,
                                         self.mview.format, self.mview.shape)

        s = b'abc'
        m = memoryview(s) 
        buf = MyBuf(m)
        buffer_support = self.get_buffer_support()
        assert buffer_support.check_readbuffer(buf)
        assert s == buffer_support.readbuffer_as_string(buf)
        assert raises(TypeError, buffer_support.writebuffer_as_string, buf)
        assert s == buffer_support.charbuffer_as_string(buf)



import time

from pypy.interpreter.error import oefmt
from pypy.interpreter.typedef import TypeDef
from pypy.interpreter.gateway import interp2app, unwrap_spec
from pypy.interpreter.baseobjspace import W_Root
from rpython.rlib.rarithmetic import r_uint, intmask, widen, LONG_BIT
from rpython.rlib import rbigint, rrandom, rstring


def descr_new__(space, w_subtype, __args__):
    w_anything = __args__.firstarg()
    x = space.allocate_instance(W_Random, w_subtype)
    x = space.interp_w(W_Random, x)
    W_Random.__init__(x, space, w_anything)
    return x


class W_Random(W_Root):
    def __init__(self, space, w_anything):
        self._rnd = rrandom.Random()
        self.seed(space, w_anything)

    def random(self, space):
        return space.newfloat(self._rnd.random())

    def seed(self, space, w_n=None):
        if w_n is None:
            w_n = space.newint(int(time.time()))
        else:
            if space.isinstance_w(w_n, space.w_int):
                w_n = space.abs(w_n)
            elif space.isinstance_w(w_n, space.w_long):
                w_n = space.abs(w_n)
            else:
                n = space.hash_w(w_n)
                w_n = space.newint(r_uint(n))
        key = []
        w_one = space.newint(1)
        w_two = space.newint(2)
        w_thirtytwo = space.newint(32)
        # 0xffffffff
        w_masklower = space.sub(space.pow(w_two, w_thirtytwo, space.w_None),
                                w_one)
        while space.is_true(w_n):
            w_chunk = space.and_(w_n, w_masklower)
            chunk = space.uint_w(w_chunk)
            key.append(chunk)
            w_n = space.rshift(w_n, w_thirtytwo)
        if not key:
            key = [r_uint(0)]
        self._rnd.init_by_array(key)

    def getstate(self, space):
        state = [None] * (rrandom.N + 1)
        for i in range(rrandom.N):
            state[i] = space.newint(widen(self._rnd.state[i]))
        state[rrandom.N] = space.newlong(self._rnd.index)
        return space.newtuple(state)

    def setstate(self, space, w_state):
        if not space.isinstance_w(w_state, space.w_tuple):
            raise oefmt(space.w_TypeError, "state vector must be tuple")
        if space.len_w(w_state) != rrandom.N + 1:
            raise oefmt(space.w_ValueError, "state vector is the wrong size")
        w_zero = space.newint(0)
        # independent of platfrom, since the below condition is only
        # true on 32 bit platforms anyway
        w_add = space.pow(space.newint(2), space.newint(32), space.w_None)
        _state = [r_uint(0)] * rrandom.N
        for i in range(rrandom.N):
            w_item = space.getitem(w_state, space.newint(i))
            if space.is_true(space.lt(w_item, w_zero)):
                w_item = space.add(w_item, w_add)
            _state[i] = space.uint_w(w_item)
        w_item = space.getitem(w_state, space.newint(rrandom.N))
        index = space.int_w(w_item)
        if index < 0 or index > rrandom.N:
            raise oefmt(space.w_ValueError, "invalid state")
        self._rnd.state = _state
        self._rnd.index = index

    def jumpahead(self, space, w_n):
        if space.isinstance_w(w_n, space.w_long):
            num = space.bigint_w(w_n)
            n = intmask(num.uintmask())
        else:
            n = space.int_w(w_n)
        self._rnd.jumpahead(n)

    @unwrap_spec(k=int)
    def getrandbits(self, space, k):
        """ getrandbits(k) -> x.  Generates a long int with k random bits. """
        if k <= 0:
            raise oefmt(space.w_ValueError,
                        "number of bits must be greater than zero")
        if k < 32: # XXX could go up to 63 bits, but let's start with this
            # fits an int, don't do the bytes-to-long-to-int dance
            r = self._rnd.genrand32()
            r >>= (32 - k)
            return space.newint(intmask(r))
        bytes = ((k - 1) // 32 + 1) * 4
        bytesarray = rstring.StringBuilder(bytes)
        for i in range(0, bytes, 4):
            r = self._rnd.genrand32()
            if k < 32:
                r >>= (32 - k)
            bytesarray.append(chr(r & r_uint(0xff)))
            bytesarray.append(chr((r >> 8) & r_uint(0xff)))
            bytesarray.append(chr((r >> 16) & r_uint(0xff)))
            bytesarray.append(chr((r >> 24) & r_uint(0xff)))
            k -= 32

        # little endian order to match bytearray assignment order
        result = rbigint.rbigint.frombytes(
            bytesarray.build(), 'little', signed=False)
        return space.newlong_from_rbigint(result)


W_Random.typedef = TypeDef("Random",
    __new__ = interp2app(descr_new__),
    random = interp2app(W_Random.random),
    seed = interp2app(W_Random.seed),
    getstate = interp2app(W_Random.getstate),
    setstate = interp2app(W_Random.setstate),
    jumpahead = interp2app(W_Random.jumpahead),
    getrandbits = interp2app(W_Random.getrandbits),
)



"""
Arguments objects.
"""
from rpython.rlib.debug import make_sure_not_resized
from rpython.rlib.objectmodel import not_rpython
from rpython.rlib import jit

from pypy.interpreter.error import OperationError, oefmt


class Arguments(object):
    """
    Collects the arguments of a function call.

    Instances should be considered immutable.

    Some parts of this class are written in a slightly convoluted style to help
    the JIT. It is really crucial to get this right, because Python's argument
    semantics are complex, but calls occur everywhere.
    """

    ###  Construction  ###

    def __init__(self, space, args_w, keywords=None, keywords_w=None,
                 w_stararg=None, w_starstararg=None, keyword_names_w=None,
                 methodcall=False):
        self.space = space
        assert isinstance(args_w, list)
        self.arguments_w = args_w
        self.keywords = keywords
        self.keywords_w = keywords_w
        self.keyword_names_w = keyword_names_w  # matches the tail of .keywords
        if keywords is not None:
            assert keywords_w is not None
            assert len(keywords_w) == len(keywords)
            assert (keyword_names_w is None or
                    len(keyword_names_w) <= len(keywords))
            make_sure_not_resized(self.keywords)
            make_sure_not_resized(self.keywords_w)

        make_sure_not_resized(self.arguments_w)
        self._combine_wrapped(w_stararg, w_starstararg)
        # a flag that specifies whether the JIT can unroll loops that operate
        # on the keywords
        self._jit_few_keywords = self.keywords is None or jit.isconstant(len(self.keywords))
        # a flag whether this is likely a method call, which doesn't change the
        # behaviour but produces better error messages
        self.methodcall = methodcall

    @not_rpython
    def __repr__(self):
        name = self.__class__.__name__
        if not self.keywords:
            return '%s(%s)' % (name, self.arguments_w,)
        else:
            return '%s(%s, %s, %s)' % (name, self.arguments_w,
                                       self.keywords, self.keywords_w)


    ###  Manipulation  ###

    @jit.look_inside_iff(lambda self: self._jit_few_keywords)
    def unpack(self): # slowish
        "Return a ([w1,w2...], {'kw':w3...}) pair."
        kwds_w = {}
        if self.keywords:
            for i in range(len(self.keywords)):
                kwds_w[self.keywords[i]] = self.keywords_w[i]
        return self.arguments_w, kwds_w

    def replace_arguments(self, args_w):
        "Return a new Arguments with a args_w as positional arguments."
        return Arguments(self.space, args_w, self.keywords, self.keywords_w,
                         keyword_names_w = self.keyword_names_w)

    def prepend(self, w_firstarg):
        "Return a new Arguments with a new argument inserted first."
        return self.replace_arguments([w_firstarg] + self.arguments_w)

    def _combine_wrapped(self, w_stararg, w_starstararg):
        "unpack the *arg and **kwd into arguments_w and keywords_w"
        if w_stararg is not None:
            self._combine_starargs_wrapped(w_stararg)
        if w_starstararg is not None:
            self._combine_starstarargs_wrapped(w_starstararg)

    def _combine_starargs_wrapped(self, w_stararg):
        # unpack the * arguments
        space = self.space
        try:
            args_w = space.fixedview(w_stararg)
        except OperationError as e:
            if (e.match(space, space.w_TypeError) and
                    not space.is_generator(w_stararg)):
                raise oefmt(space.w_TypeError,
                            "argument after * must be an iterable, not %T",
                            w_stararg)
            raise
        self.arguments_w = self.arguments_w + args_w

    def _combine_starstarargs_wrapped(self, w_starstararg):
        # unpack the ** arguments
        space = self.space
        keywords, values_w = space.view_as_kwargs(w_starstararg)
        if keywords is not None: # this path also taken for empty dicts
            if self.keywords is None:
                self.keywords = keywords
                self.keywords_w = values_w
            else:
                _check_not_duplicate_kwargs(
                    self.space, self.keywords, keywords, values_w)
                self.keywords = self.keywords + keywords
                self.keywords_w = self.keywords_w + values_w
            return
        is_dict = False
        if space.isinstance_w(w_starstararg, space.w_dict):
            is_dict = True
            keys_w = space.unpackiterable(w_starstararg)
        else:
            try:
                w_keys = space.call_method(w_starstararg, "keys")
            except OperationError as e:
                if e.match(space, space.w_AttributeError):
                    raise oefmt(space.w_TypeError,
                                "argument after ** must be a mapping, not %T",
                                w_starstararg)
                raise
            keys_w = space.unpackiterable(w_keys)
        keywords_w = [None] * len(keys_w)
        keywords = [None] * len(keys_w)
        _do_combine_starstarargs_wrapped(
            space, keys_w, w_starstararg, keywords, keywords_w, self.keywords,
            is_dict)
        self.keyword_names_w = keys_w
        if self.keywords is None:
            self.keywords = keywords
            self.keywords_w = keywords_w
        else:
            self.keywords = self.keywords + keywords
            self.keywords_w = self.keywords_w + keywords_w


    def fixedunpack(self, argcount):
        """The simplest argument parsing: get the 'argcount' arguments,
        or raise a real ValueError if the length is wrong."""
        if self.keywords:
            raise ValueError("no keyword arguments expected")
        if len(self.arguments_w) > argcount:
            raise ValueError("too many arguments (%d expected)" % argcount)
        elif len(self.arguments_w) < argcount:
            raise ValueError("not enough arguments (%d expected)" % argcount)
        return self.arguments_w

    def firstarg(self):
        "Return the first argument for inspection."
        if self.arguments_w:
            return self.arguments_w[0]
        return None

    ###  Parsing for function calls  ###

    @jit.unroll_safe
    def _match_signature(self, w_firstarg, scope_w, signature, defaults_w=None,
                         blindargs=0):
        """Parse args and kwargs according to the signature of a code object,
        or raise an ArgErr in case of failure.
        """
        #   w_firstarg = a first argument to be inserted (e.g. self) or None
        #   args_w = list of the normal actual parameters, wrapped
        #   scope_w = resulting list of wrapped values
        #

        # some comments about the JIT: it assumes that signature is a constant,
        # so all values coming from there can be assumed constant. It assumes
        # that the length of the defaults_w does not vary too much.
        co_argcount = signature.num_argnames() # expected formal arguments, without */**

        # put the special w_firstarg into the scope, if it exists
        if w_firstarg is not None:
            upfront = 1
            if co_argcount > 0:
                scope_w[0] = w_firstarg
        else:
            upfront = 0

        args_w = self.arguments_w
        num_args = len(args_w)
        avail = num_args + upfront

        keywords = self.keywords
        num_kwds = 0
        if keywords is not None:
            num_kwds = len(keywords)


        # put as many positional input arguments into place as available
        input_argcount = upfront
        if input_argcount < co_argcount:
            take = min(num_args, co_argcount - upfront)

            # letting the JIT unroll this loop is safe, because take is always
            # smaller than co_argcount
            for i in range(take):
                scope_w[i + input_argcount] = args_w[i]
            input_argcount += take

        # collect extra positional arguments into the *vararg
        if signature.has_vararg():
            args_left = co_argcount - upfront
            if args_left < 0:  # check required by rpython
                starargs_w = [w_firstarg]
                if num_args:
                    starargs_w = starargs_w + args_w
            elif num_args > args_left:
                if args_left == 0:
                    starargs_w = args_w # don't need to copy if its the whole list
                else:
                    starargs_w = args_w[args_left:]
            else:
                starargs_w = []
            scope_w[co_argcount] = self.space.newtuple(starargs_w)
        elif avail > co_argcount:
            raise self.argerrcount(avail, num_kwds, signature, defaults_w, 0)

        # if a **kwargs argument is needed, create the dict
        w_kwds = None
        if signature.has_kwarg():
            w_kwds = self.space.newdict(kwargs=True)
            scope_w[co_argcount + signature.has_vararg()] = w_kwds

        # handle keyword arguments
        num_remainingkwds = 0
        keywords_w = self.keywords_w
        kwds_mapping = None
        if num_kwds:
            # kwds_mapping maps target indexes in the scope (minus input_argcount)
            # to positions in the keywords_w list
            kwds_mapping = [0] * (co_argcount - input_argcount)
            # initialize manually, for the JIT :-(
            for i in range(len(kwds_mapping)):
                kwds_mapping[i] = -1
            # match the keywords given at the call site to the argument names
            # the called function takes
            # this function must not take a scope_w, to make the scope not
            # escape
            num_remainingkwds = _match_keywords(
                    signature, blindargs, input_argcount, keywords,
                    kwds_mapping, self._jit_few_keywords)
            if num_remainingkwds:
                if w_kwds is not None:
                    # collect extra keyword arguments into the **kwarg
                    _collect_keyword_args(
                            self.space, keywords, keywords_w, w_kwds,
                            kwds_mapping, self.keyword_names_w, self._jit_few_keywords)
                else:
                    if co_argcount == 0:
                        raise self.argerrcount(avail, num_kwds, signature, defaults_w, 0)
                    raise ArgErrUnknownKwds(self.space, num_remainingkwds, keywords,
                                            kwds_mapping, self.keyword_names_w)

        # check for missing arguments and fill them from the kwds,
        # or with defaults, if available
        missing = 0
        if input_argcount < co_argcount:
            def_first = co_argcount - (0 if defaults_w is None else len(defaults_w))
            j = 0
            kwds_index = -1
            for i in range(input_argcount, co_argcount):
                if kwds_mapping is not None:
                    kwds_index = kwds_mapping[j]
                    j += 1
                    if kwds_index >= 0:
                        scope_w[i] = keywords_w[kwds_index]
                        continue
                defnum = i - def_first
                if defnum >= 0:
                    scope_w[i] = defaults_w[defnum]
                else:
                    missing += 1
            if missing:
                raise self.argerrcount(avail, num_kwds, signature, defaults_w, missing)

    def argerrcount(self, *args):
        if self.methodcall:
            return ArgErrCountMethod(*args)
        return ArgErrCount(*args)

    def parse_into_scope(self, w_firstarg,
                         scope_w, fnname, signature, defaults_w=None):
        """Parse args and kwargs to initialize a frame
        according to the signature of code object.
        Store the argumentvalues into scope_w.
        scope_w must be big enough for signature.
        """
        try:
            self._match_signature(w_firstarg,
                                  scope_w, signature, defaults_w, 0)
        except ArgErr as e:
            raise oefmt(self.space.w_TypeError, "%s() %s", fnname, e.getmsg())
        return signature.scope_length()

    def _parse(self, w_firstarg, signature, defaults_w, blindargs=0):
        """Parse args and kwargs according to the signature of a code object,
        or raise an ArgErr in case of failure.
        """
        scopelen = signature.scope_length()
        scope_w = [None] * scopelen
        self._match_signature(w_firstarg, scope_w, signature, defaults_w,
                              blindargs)
        return scope_w


    def parse_obj(self, w_firstarg,
                  fnname, signature, defaults_w=None, blindargs=0):
        """Parse args and kwargs to initialize a frame
        according to the signature of code object.
        """
        try:
            return self._parse(w_firstarg, signature, defaults_w, blindargs)
        except ArgErr as e:
            raise oefmt(self.space.w_TypeError, "%s() %s", fnname, e.getmsg())

    @staticmethod
    def frompacked(space, w_args=None, w_kwds=None):
        """Convenience static method to build an Arguments
           from a wrapped sequence and a wrapped dictionary."""
        return Arguments(space, [], w_stararg=w_args, w_starstararg=w_kwds)

    def topacked(self):
        """Express the Argument object as a pair of wrapped w_args, w_kwds."""
        space = self.space
        w_args = space.newtuple(self.arguments_w)
        w_kwds = space.newdict()
        if self.keywords is not None:
            limit = len(self.keywords)
            if self.keyword_names_w is not None:
                limit -= len(self.keyword_names_w)
            for i in range(len(self.keywords)):
                if i < limit:
                    key = self.keywords[i]
                    space.setitem_str(w_kwds, key, self.keywords_w[i])
                else:
                    w_key = self.keyword_names_w[i - limit]
                    space.setitem(w_kwds, w_key, self.keywords_w[i])
        return w_args, w_kwds

# JIT helper functions
# these functions contain functionality that the JIT is not always supposed to
# look at. They should not get a self arguments, which makes the amount of
# arguments annoying :-(

@jit.look_inside_iff(lambda space, existingkeywords, keywords, keywords_w:
        jit.isconstant(len(keywords) and
        jit.isconstant(existingkeywords)))
def _check_not_duplicate_kwargs(space, existingkeywords, keywords, keywords_w):
    # looks quadratic, but the JIT should remove all of it nicely.
    # Also, all the lists should be small
    for key in keywords:
        for otherkey in existingkeywords:
            if otherkey == key:
                raise oefmt(space.w_TypeError,
                            "got multiple values for keyword argument '%s'",
                            key)

def _do_combine_starstarargs_wrapped(space, keys_w, w_starstararg, keywords,
        keywords_w, existingkeywords, is_dict):
    i = 0
    for w_key in keys_w:
        try:
            key = space.text_w(w_key)
        except OperationError as e:
            if e.match(space, space.w_TypeError):
                raise oefmt(space.w_TypeError, "keywords must be strings")
            if e.match(space, space.w_UnicodeEncodeError):
                # Allow this to pass through
                key = None
            else:
                raise
        else:
            if existingkeywords and key in existingkeywords:
                raise oefmt(space.w_TypeError,
                            "got multiple values for keyword argument '%s'",
                            key)
        keywords[i] = key
        if is_dict:
            # issue 2435: bug-to-bug compatibility with cpython. for a subclass of
            # dict, just ignore the __getitem__ and access the underlying dict
            # directly
            from pypy.objspace.descroperation import dict_getitem
            w_descr = dict_getitem(space)
            w_value = space.get_and_call_function(w_descr, w_starstararg, w_key)
        else:
            w_value = space.getitem(w_starstararg, w_key)
        keywords_w[i] = w_value
        i += 1

@jit.look_inside_iff(
    lambda signature, blindargs, input_argcount,
           keywords, kwds_mapping, jiton: jiton)
def _match_keywords(signature, blindargs, input_argcount,
                    keywords, kwds_mapping, _):
    # letting JIT unroll the loop is *only* safe if the callsite didn't
    # use **args because num_kwds can be arbitrarily large otherwise.
    num_kwds = num_remainingkwds = len(keywords)
    for i in range(num_kwds):
        name = keywords[i]
        # If name was not encoded as a string, it could be None. In that
        # case, it's definitely not going to be in the signature.
        if name is None:
            continue
        j = signature.find_argname(name)
        # if j == -1 nothing happens, because j < input_argcount and
        # blindargs > j
        if j < input_argcount:
            # check that no keyword argument conflicts with these. note
            # that for this purpose we ignore the first blindargs,
            # which were put into place by prepend().  This way,
            # keywords do not conflict with the hidden extra argument
            # bound by methods.
            if blindargs <= j:
                raise ArgErrMultipleValues(name)
        else:
            kwds_mapping[j - input_argcount] = i # map to the right index
            num_remainingkwds -= 1
    return num_remainingkwds

@jit.look_inside_iff(
    lambda space, keywords, keywords_w, w_kwds, kwds_mapping,
        keyword_names_w, jiton: jiton)
def _collect_keyword_args(space, keywords, keywords_w, w_kwds, kwds_mapping,
                          keyword_names_w, _):
    limit = len(keywords)
    if keyword_names_w is not None:
        limit -= len(keyword_names_w)
    for i in range(len(keywords)):
        # again a dangerous-looking loop that either the JIT unrolls
        # or that is not too bad, because len(kwds_mapping) is small
        for j in kwds_mapping:
            if i == j:
                break
        else:
            if i < limit:
                space.setitem_str(w_kwds, keywords[i], keywords_w[i])
            else:
                w_key = keyword_names_w[i - limit]
                space.setitem(w_kwds, w_key, keywords_w[i])

#
# ArgErr family of exceptions raised in case of argument mismatch.
# We try to give error messages following CPython's, which are very informative.
#

class ArgErr(Exception):

    def getmsg(self):
        raise NotImplementedError

class ArgErrCount(ArgErr):

    def __init__(self, got_nargs, nkwds, signature,
                 defaults_w, missing_args):
        self.signature = signature

        self.num_defaults = 0 if defaults_w is None else len(defaults_w)
        self.missing_args = missing_args
        self.num_args = got_nargs
        self.num_kwds = nkwds

    def getmsg(self):
        n = self.signature.num_argnames()
        if n == 0:
            msg = "takes no arguments (%d given)" % (
                self.num_args + self.num_kwds)
        else:
            defcount = self.num_defaults
            has_kwarg = self.signature.has_kwarg()
            num_args = self.num_args
            num_kwds = self.num_kwds
            if defcount == 0 and not self.signature.has_vararg():
                msg1 = "exactly"
                if not has_kwarg:
                    num_args += num_kwds
                    num_kwds = 0
            elif not self.missing_args:
                msg1 = "at most"
            else:
                msg1 = "at least"
                has_kwarg = False
                n -= defcount
            if n == 1:
                plural = ""
            else:
                plural = "s"
            if has_kwarg or num_kwds > 0:
                msg2 = " non-keyword"
            else:
                msg2 = ""
            msg = "takes %s %d%s argument%s (%d given)" % (
                msg1,
                n,
                msg2,
                plural,
                num_args)
        return msg

class ArgErrCountMethod(ArgErrCount):
    """ A subclass of ArgErrCount that is used if the argument matching is done
    as part of a method call, in which case more information is added to the
    error message, if the cause of the error is likely a forgotten `self`
    argument.
    """

    def getmsg(self):
        msg = ArgErrCount.getmsg(self)
        n = self.signature.num_argnames()
        if (self.num_args == n + 1 and
                (n == 0 or self.signature.argnames[0] != "self")):
            msg += ". Did you forget 'self' in the function definition?"
        return msg


class ArgErrMultipleValues(ArgErr):

    def __init__(self, argname):
        self.argname = argname

    def getmsg(self):
        msg = "got multiple values for keyword argument '%s'" % (
            self.argname)
        return msg

class ArgErrUnknownKwds(ArgErr):

    def __init__(self, space, num_remainingkwds, keywords, kwds_mapping,
                 keyword_names_w):
        name = ''
        self.num_kwds = num_remainingkwds
        if num_remainingkwds == 1:
            for i in range(len(keywords)):
                if i not in kwds_mapping:
                    name = '?'
                    # We'll assume it's unicode. Encode it.
                    # Careful, I *think* it should not be possible to
                    # get an IndexError here but you never know.
                    try:
                        if keyword_names_w is None:
                            raise IndexError
                        # note: negative-based indexing from the end
                        w_name = keyword_names_w[i - len(keywords)]
                    except IndexError:
                        if keywords is None:
                            name = '?'
                        else:
                            name = keywords[i]
                    else:
                        w_enc = space.newtext(space.sys.defaultencoding)
                        w_err = space.newtext("replace")
                        w_name = space.call_method(w_name, "encode", w_enc,
                                                   w_err)
                        name = space.text_w(w_name)
                    break
        self.kwd_name = name

    def getmsg(self):
        if self.num_kwds == 1:
            msg = "got an unexpected keyword argument '%s'" % (
                self.kwd_name)
        else:
            msg = "got %d unexpected keyword arguments" % (
                self.num_kwds)
        return msg



import pytest
from pypy.interpreter.pyparser import pytokenizer
from pypy.interpreter.pyparser.parser import Token
from pypy.interpreter.pyparser.pygram import tokens
from pypy.interpreter.pyparser.error import TokenError

def tokenize(s):
    return pytokenizer.generate_tokens(s.splitlines(True) + ["\n"], 0)

def check_token_error(s, msg=None, pos=-1, line=-1):
    error = pytest.raises(TokenError, tokenize, s)
    if msg is not None:
        assert error.value.msg == msg
    if pos != -1:
        assert error.value.offset == pos
    if line != -1:
        assert error.value.lineno == line


class TestTokenizer(object):

    def test_simple(self):
        line = "a+1"
        tks = tokenize(line)
        assert tks == [
            Token(tokens.NAME, 'a', 1, 0, line),
            Token(tokens.PLUS, '+', 1, 1, line),
            Token(tokens.NUMBER, '1', 1, 2, line),
            Token(tokens.NEWLINE, '', 2, 0, '\n'),
            Token(tokens.NEWLINE, '', 2, 0, '\n'),
            Token(tokens.ENDMARKER, '', 2, 0, ''),
            ]

    def test_error_parenthesis(self):
        for paren in "([{":
            check_token_error(paren + "1 + 2",
                              "parenthesis is never closed",
                              1)

        for paren in ")]}":
            check_token_error("1 + 2" + paren,
                              "unmatched '%s'" % (paren, ),
                              6)

        for i, opening in enumerate("([{"):
            for j, closing in enumerate(")]}"):
                if i == j:
                    continue
                check_token_error(opening + "1\n" + closing,
                        "closing parenthesis '%s' does not match opening parenthesis '%s' on line 1" % (closing, opening),
                        pos=1, line=2)
                check_token_error(opening + "1" + closing,
                        "closing parenthesis '%s' does not match opening parenthesis '%s'" % (closing, opening),
                        pos=3, line=1)
                check_token_error(opening + closing,
                        "closing parenthesis '%s' does not match opening parenthesis '%s'" % (closing, opening),
                        pos=2, line=1)


    def test_unknown_char(self):
        check_token_error("?", "Unknown character", 1)

    def test_eol_string(self):
        check_token_error("x = 'a", pos=5, line=1)

    def test_eof_triple_quoted(self):
        check_token_error("'''", pos=1, line=1)



from pypy.interpreter.gateway import unwrap_spec
from rpython.rlib.rstring import StringBuilder

MAXLINESIZE = 76

# ____________________________________________________________

def hexval(c):
    if c <= '9':
        return ord(c) - ord('0')
    elif c <= 'F':
        return ord(c) - (ord('A') - 10)
    else:
        return ord(c) - (ord('a') - 10)
hexval._always_inline_ = True

@unwrap_spec(data='bufferstr', header=int)
def a2b_qp(space, data, header=0):
    "Decode a string of qp-encoded data."

    # We allocate the output same size as input, this is overkill.
    odata = StringBuilder(len(data))
    inp = 0

    while inp < len(data):
        c = data[inp]
        inp += 1
        if c == '=':
            if inp >= len(data):
                break
            # Soft line breaks
            c = data[inp]
            if c == '\n' or c == '\r':
                if c != '\n':
                    while inp < len(data) and data[inp] != '\n':
                        inp += 1
                inp += 1   # may go beyond len(data)
            elif c == '=':
                # broken case from broken python qp
                odata.append('=')
                inp += 1
            elif (inp + 1 < len(data) and
                  ('A' <= c <= 'F' or
                   'a' <= c <= 'f' or
                   '0' <= c <= '9') and
                  ('A' <= data[inp+1] <= 'F' or
                   'a' <= data[inp+1] <= 'f' or
                   '0' <= data[inp+1] <= '9')):
                # hexval
                ch = chr(hexval(c) << 4 | hexval(data[inp+1]))
                inp += 2
                odata.append(ch)
            else:
                odata.append('=')
        else:
            if header and c == '_':
                c = ' '
            odata.append(c)
    return space.newbytes(odata.build())

# ____________________________________________________________

class StringBuilderWithOneCharCancellable(object):

    def __init__(self, crlf, initial):
        self.crlf = crlf
        self.builder = StringBuilder(initial)
        self.pending = -1

    def _flush(self):
        if self.pending >= 0:
            self.builder.append(chr(self.pending))
            self.pending = -1
    _flush._always_inline_ = True

    def append(self, c):
        self._flush()
        self.pending = ord(c)

    def newline(self):
        self._flush()
        if self.crlf: self.builder.append('\r')
        self.pending = ord('\n')

    def to_hex(self, c):
        self._flush()
        uvalue = ord(c)
        self.builder.append("0123456789ABCDEF"[uvalue >> 4])
        self.builder.append("0123456789ABCDEF"[uvalue & 0xf])

    def build(self):
        self._flush()
        return self.builder.build()

@unwrap_spec(data='bufferstr', quotetabs=int, istext=int, header=int)
def b2a_qp(space, data, quotetabs=0, istext=1, header=0):
    """Encode a string using quoted-printable encoding.

On encoding, when istext is set, newlines are not encoded, and white
space at end of lines is.  When istext is not set, \\r and \\n (CR/LF) are
both encoded.  When quotetabs is set, space and tabs are encoded."""

    # See if this string is using CRLF line ends
    lf = data.find('\n')
    crlf = lf > 0 and data[lf-1] == '\r'

    # We allocate the output initially the same size as input;
    # it may need resizing.
    odata = StringBuilderWithOneCharCancellable(crlf, len(data))
    inp = 0
    linelen = 0

    while inp < len(data):
        c = data[inp]
        if (c > '~' or
            c == '=' or
            (header and c == '_') or
            (c == '.' and linelen == 0 and (inp + 1 == len(data) or
                                            data[inp+1] in '\n\r\x00')) or
            (not istext and (c == '\r' or c == '\n')) or
            ((c == '\t' or c == ' ') and (inp + 1 == len(data))) or
            (c <= ' ' and c != '\r' and c != '\n' and
             (quotetabs or (c != '\t' and c != ' ')))):
            linelen += 3
            if linelen >= MAXLINESIZE:
                odata.append('=')
                odata.newline()
                linelen = 3
            odata.append('=')
            odata.to_hex(c)
            inp += 1
        else:
            if (istext and
                (c == '\n' or (inp+1 < len(data) and c == '\r' and
                               data[inp+1] == '\n'))):
                # Protect against whitespace on end of line
                pendingnum = odata.pending
                if pendingnum == ord(' ') or pendingnum == ord('\t'):
                    odata.pending = ord('=')
                    odata.to_hex(chr(pendingnum))

                linelen = 0
                odata.newline()
                if c == '\r':
                    inp += 2
                else:
                    inp += 1
            else:
                if (inp + 1 < len(data) and
                    data[inp+1] != '\n' and
                    (linelen + 1) >= MAXLINESIZE):
                    odata.append('=')
                    odata.newline()
                    linelen = 0

                linelen += 1
                if header and c == ' ':
                    c = '_'
                odata.append(c)
                inp += 1

    return space.newbytes(odata.build())



"""
A simple deeply recursive target. 

The target below specifies None as the argument types list.
This is a case treated specially in driver.py . If the list
of input types is empty, it is meant to be a list of strings,
actually implementing argv of the executable.
"""

import os, sys

def debug(msg): 
    os.write(2, "debug: " + msg + '\n')

# __________  Entry point  __________

def ackermann(x, y):
    if x == 0:
        return y + 1
    if y == 0:
        return ackermann(x - 1, 1)
    return ackermann(x - 1, ackermann(x, y - 1))

def entry_point(argv):
    debug(str(ackermann(3, 12)) + "\n")
    return 0

# _____ Define and setup target ___

def target(*args):
    return entry_point, None



# Package initialisation
from pypy.interpreter.mixedmodule import MixedModule
import errno

class Module(MixedModule):
    """This module makes available standard errno system symbols.

    The value of each symbol is the corresponding integer value,
    e.g., on most systems, errno.ENOENT equals the integer 2.

    The dictionary errno.errorcode maps numeric codes to symbol names,
    e.g., errno.errorcode[2] could be the string 'ENOENT'.

    Symbols that are not relevant to the underlying system are not defined.

    To map error codes to error messages, use the function os.strerror(),
    e.g. os.strerror(2) could return 'No such file or directory'."""

    appleveldefs = {}
    interpleveldefs = {"errorcode": "interp_errno.get_errorcode(space)"}
    
for name in dir(errno):
    if name.startswith('__') or name in Module.interpleveldefs:
        continue
    Module.interpleveldefs[name] = ("space.newint(%s)" %
                                    (getattr(errno, name), ))



import py
from pypy.module.cpyext.test.test_cpyext import AppTestCpythonExtensionBase
from pypy.module.cpyext.test.test_api import BaseApiTest
from pypy.module.cpyext.pyobject import make_ref


class AppTestBorrow(AppTestCpythonExtensionBase):
    def test_tuple_borrowing(self):
        module = self.import_extension('foo', [
            ("test_borrowing", "METH_NOARGS",
             """
                PyObject *t = PyTuple_New(1);
                PyObject *f = PyFloat_FromDouble(42.0);
                PyObject *g = NULL;
                printf("Refcnt1: %ld\\n", f->ob_refcnt);
                PyTuple_SetItem(t, 0, f); // steals reference
                printf("Refcnt2: %ld\\n", f->ob_refcnt);
                f = PyTuple_GetItem(t, 0); // borrows reference
                printf("Refcnt3: %ld\\n", f->ob_refcnt);
                g = PyTuple_GetItem(t, 0); // borrows reference again
                printf("Refcnt4: %ld\\n", f->ob_refcnt);
                printf("COMPARE: %i\\n", f == g);
                fflush(stdout);
                Py_DECREF(t);
                Py_RETURN_TRUE;
             """),
            ])
        assert module.test_borrowing() # the test should not leak

    def test_borrow_destroy(self):
        module = self.import_extension('foo', [
            ("test_borrow_destroy", "METH_NOARGS",
             """
                PyObject *i = PyInt_FromLong(42);
                PyObject *j;
                PyObject *t1 = PyTuple_Pack(1, i);
                PyObject *t2 = PyTuple_Pack(1, i);
                Py_DECREF(i);

                i = PyTuple_GetItem(t1, 0);
                PyTuple_GetItem(t2, 0);
                Py_DECREF(t2);

                j = PyInt_FromLong(PyInt_AsLong(i));
                Py_DECREF(t1);
                return j;
             """),
            ])
        assert module.test_borrow_destroy() == 42

    def test_double_borrow(self):
        if self.runappdirect:
            py.test.xfail('segfault')
        module = self.import_extension('foo', [
            ("run", "METH_NOARGS",
             """
                PyObject *t = PyTuple_New(1);
                PyObject *s = PyRun_String("set()", Py_eval_input,
                                           Py_None, Py_None);
                PyObject *w = PyWeakref_NewRef(s, Py_None);
                PyTuple_SetItem(t, 0, s);
                PyTuple_GetItem(t, 0);
                PyTuple_GetItem(t, 0);
                Py_DECREF(t);
                return w;
             """),
            ])
        wr = module.run()
        # check that the set() object was deallocated
        self.debug_collect()
        assert wr() is None



import py
from pypy.objspace.std.test.test_dictmultiobject import FakeSpace, W_DictObject
from pypy.objspace.std.kwargsdict import *

space = FakeSpace()
strategy = KwargsDictStrategy(space)

def test_create():
    keys = ["a", "b", "c"]
    values = [1, 2, 3]
    storage = strategy.erase((keys, values))
    d = W_DictObject(space, strategy, storage)
    assert d.getitem_str("a") == 1
    assert d.getitem_str("b") == 2
    assert d.getitem_str("c") == 3
    assert d.getitem(space.wrap("a")) == 1
    assert d.getitem(space.wrap("b")) == 2
    assert d.getitem(space.wrap("c")) == 3
    assert d.w_keys() == keys
    assert d.values() == values

def test_set_existing():
    keys = ["a", "b", "c"]
    values = [1, 2, 3]
    storage = strategy.erase((keys, values))
    d = W_DictObject(space, strategy, storage)
    assert d.getitem_str("a") == 1
    assert d.getitem_str("b") == 2
    assert d.getitem_str("c") == 3
    assert d.setitem_str("a", 4) is None
    assert d.getitem_str("a") == 4
    assert d.getitem_str("b") == 2
    assert d.getitem_str("c") == 3
    assert d.setitem_str("b", 5) is None
    assert d.getitem_str("a") == 4
    assert d.getitem_str("b") == 5
    assert d.getitem_str("c") == 3
    assert d.setitem_str("c", 6) is None
    assert d.getitem_str("a") == 4
    assert d.getitem_str("b") == 5
    assert d.getitem_str("c") == 6
    assert d.getitem(space.wrap("a")) == 4
    assert d.getitem(space.wrap("b")) == 5
    assert d.getitem(space.wrap("c")) == 6
    assert d.w_keys() == keys
    assert d.values() == values
    assert keys == ["a", "b", "c"]
    assert values == [4, 5, 6]


def test_set_new():
    keys = ["a", "b", "c"]
    values = [1, 2, 3]
    storage = strategy.erase((keys, values))
    d = W_DictObject(space, strategy, storage)
    assert d.getitem_str("a") == 1
    assert d.getitem_str("b") == 2
    assert d.getitem_str("c") == 3
    assert d.getitem_str("d") is None
    assert d.setitem_str("d", 4) is None
    assert d.getitem_str("a") == 1
    assert d.getitem_str("b") == 2
    assert d.getitem_str("c") == 3
    assert d.getitem_str("d") == 4
    assert d.w_keys() == keys
    assert d.values() == values
    assert keys == ["a", "b", "c", "d"]
    assert values == [1, 2, 3, 4]

def test_limit_size():
    storage = strategy.get_empty_storage()
    d = W_DictObject(space, strategy, storage)
    for i in range(100):
        assert d.setitem_str("d%s" % i, 4) is None
    assert d.get_strategy() is not strategy
    assert "BytesDictStrategy" == d.get_strategy().__class__.__name__

def test_keys_doesnt_wrap():
    space = FakeSpace()
    space.newlist = None
    strategy = KwargsDictStrategy(space)
    keys = ["a", "b", "c"]
    values = [1, 2, 3]
    storage = strategy.erase((keys, values))
    d = W_DictObject(space, strategy, storage)
    w_l = d.w_keys() # does not crash

def test_view_as_kwargs():
    from pypy.objspace.std.dictmultiobject import EmptyDictStrategy
    strategy = KwargsDictStrategy(space)
    keys = ["a", "b", "c"]
    values = [1, 2, 3]
    storage = strategy.erase((keys, values))
    d = W_DictObject(space, strategy, storage)
    assert space.view_as_kwargs(d) == (keys, values)

    strategy = EmptyDictStrategy(space)
    storage = strategy.get_empty_storage()
    d = W_DictObject(space, strategy, storage)
    assert space.view_as_kwargs(d) == ([], [])

def test_from_empty_to_kwargs():
    strategy = EmptyKwargsDictStrategy(space)
    storage = strategy.get_empty_storage()
    d = W_DictObject(space, strategy, storage)
    d.setitem_str("a", 3)
    assert isinstance(d.get_strategy(), KwargsDictStrategy)


from pypy.objspace.std.test.test_dictmultiobject import BaseTestRDictImplementation, BaseTestDevolvedDictImplementation
def get_impl(self):
    storage = strategy.erase(([], []))
    return W_DictObject(space, strategy, storage)

class TestKwargsDictImplementation(BaseTestRDictImplementation):
    StrategyClass = KwargsDictStrategy
    get_impl = get_impl
    def test_delitem(self):
        pass # delitem devolves for now

    def test_setdefault_fast(self):
        pass # not based on hashing at all

class TestDevolvedKwargsDictImplementation(BaseTestDevolvedDictImplementation):
    get_impl = get_impl
    StrategyClass = KwargsDictStrategy

    def test_setdefault_fast(self):
        pass # not based on hashing at all


class AppTestKwargsDictStrategy(object):
    def setup_class(cls):
        if cls.runappdirect:
            py.test.skip("__repr__ doesn't work on appdirect")

    def w_get_strategy(self, obj):
        import __pypy__
        r = __pypy__.internal_repr(obj)
        return r[r.find("(") + 1: r.find(")")]

    def test_create(self):
        def f(**args):
            return args
        d = f(a=1)
        assert "KwargsDictStrategy" in self.get_strategy(d)
        d = f()
        assert "EmptyKwargsDictStrategy" in self.get_strategy(d)

    def test_iterator(self):
        def f(**args):
            return args

        assert dict.fromkeys(f(a=2, b=3)) == {"a": None, "b": None}
        assert sorted(f(a=2, b=3).itervalues()) == [2, 3]

    def test_setdefault(self):
        def f(**args):
            return args
        d = f(a=1, b=2)
        a = d.setdefault("a", 0)
        assert a == 1
        a = d.setdefault("b", 0)
        assert a == 2
        a = d.setdefault("c", 3)
        assert a == 3
        assert "KwargsDictStrategy" in self.get_strategy(d)

    def test_iteritems_bug(self):
        def f(**args):
            return args

        d = f(a=2, b=3, c=4)
        for key, value in d.iteritems():
            None in d

    def test_popitem_bug(self):
        def f(**args):
            return args
        d = f(a=1, b=2)
        x = d.popitem()
        assert x == ('b', 2)
        x = d.popitem()
        assert x == ('a', 1)
        raises(KeyError, d.popitem)




""" A small helper module for profiling JIT
"""

import time
from rpython.rlib.debug import debug_print, debug_start, debug_stop
from rpython.rlib.debug import have_debug_prints
from rpython.jit.metainterp.jitexc import JitException
from rpython.rlib.jit import Counters


JITPROF_LINES = Counters.ncounters + 1 + 1
# one for TOTAL, 1 for calls, update if needed
_CPU_LINES = 4       # the last 4 lines are stored on the cpu

class BaseProfiler(object):
    pass

class EmptyProfiler(BaseProfiler):
    initialized = True

    def start(self):
        pass

    def finish(self):
        pass

    def start_tracing(self):
        pass

    def end_tracing(self):
        pass

    def start_backend(self):
        pass

    def end_backend(self):
        pass

    def count(self, kind, inc=1):
        pass

    def count_ops(self, opnum, kind=Counters.OPS):
        pass

    def get_counter(self, num):
        return 0

    def get_times(self, num):
        return 0.0

class Profiler(BaseProfiler):
    initialized = False
    timer = staticmethod(time.time)
    starttime = 0
    t1 = 0
    times = None
    counters = None
    calls = 0
    current = None
    cpu = None

    def start(self):
        self.starttime = self.timer()
        self.t1 = self.starttime
        self.times = [0, 0]
        self.counters = [0] * (Counters.ncounters - _CPU_LINES)
        self.calls = 0
        self.current = []

    def finish(self):
        self.tk = self.timer()
        self.print_stats()

    def _start(self, event):
        t0 = self.t1
        self.t1 = self.timer()
        if self.current:
            self.times[self.current[-1]] += self.t1 - t0
        self.counters[event] += 1
        self.current.append(event)

    def _end(self, event):
        t0 = self.t1
        self.t1 = self.timer()
        if not self.current:
            debug_print("BROKEN PROFILER DATA!")
            return
        ev1 = self.current.pop()
        if ev1 != event:
            debug_print("BROKEN PROFILER DATA!")
            return
        self.times[ev1] += self.t1 - t0

    def start_tracing(self):   self._start(Counters.TRACING)
    def end_tracing(self):     self._end  (Counters.TRACING)

    def start_backend(self):   self._start(Counters.BACKEND)
    def end_backend(self):     self._end  (Counters.BACKEND)

    def count(self, kind, inc=1):
        self.counters[kind] += inc

    def get_counter(self, num):
        if num == Counters.TOTAL_COMPILED_LOOPS:
            return self.cpu.tracker.total_compiled_loops
        elif num == Counters.TOTAL_COMPILED_BRIDGES:
            return self.cpu.tracker.total_compiled_bridges
        elif num == Counters.TOTAL_FREED_LOOPS:
            return self.cpu.tracker.total_freed_loops
        elif num == Counters.TOTAL_FREED_BRIDGES:
            return self.cpu.tracker.total_freed_bridges
        return self.counters[num]

    def get_times(self, num):
        return self.times[num]

    def count_ops(self, opnum, kind=Counters.OPS):
        from rpython.jit.metainterp.resoperation import OpHelpers
        self.counters[kind] += 1
        if OpHelpers.is_call(opnum) and kind == Counters.RECORDED_OPS:
            self.calls += 1

    def print_stats(self):
        debug_start("jit-summary")
        if have_debug_prints():
            self._print_stats()
        debug_stop("jit-summary")

    def _print_stats(self):
        cnt = self.counters
        tim = self.times
        calls = self.calls
        self._print_line_time("Tracing", cnt[Counters.TRACING],
                              tim[Counters.TRACING])
        self._print_line_time("Backend", cnt[Counters.BACKEND],
                              tim[Counters.BACKEND])
        line = "TOTAL:      \t\t%f" % (self.tk - self.starttime, )
        debug_print(line)
        self._print_intline("ops", cnt[Counters.OPS])
        self._print_intline("heapcached ops", cnt[Counters.HEAPCACHED_OPS])
        self._print_intline("recorded ops", cnt[Counters.RECORDED_OPS])
        self._print_intline("  calls", calls)
        self._print_intline("guards", cnt[Counters.GUARDS])
        self._print_intline("opt ops", cnt[Counters.OPT_OPS])
        self._print_intline("opt guards", cnt[Counters.OPT_GUARDS])
        self._print_intline("opt guards shared", cnt[Counters.OPT_GUARDS_SHARED])
        self._print_intline("forcings", cnt[Counters.OPT_FORCINGS])
        self._print_intline("abort: trace too long",
                            cnt[Counters.ABORT_TOO_LONG])
        self._print_intline("abort: compiling", cnt[Counters.ABORT_BRIDGE])
        self._print_intline("abort: vable escape", cnt[Counters.ABORT_ESCAPE])
        self._print_intline("abort: bad loop", cnt[Counters.ABORT_BAD_LOOP])
        self._print_intline("abort: force quasi-immut",
                            cnt[Counters.ABORT_FORCE_QUASIIMMUT])
        self._print_intline("abort: segmenting trace",
                            cnt[Counters.ABORT_SEGMENTED_TRACE])
        self._print_intline("virtualizables forced",
                            cnt[Counters.FORCE_VIRTUALIZABLES])
        self._print_intline("nvirtuals", cnt[Counters.NVIRTUALS])
        self._print_intline("nvholes", cnt[Counters.NVHOLES])
        self._print_intline("nvreused", cnt[Counters.NVREUSED])
        self._print_intline("vecopt tried", cnt[Counters.OPT_VECTORIZE_TRY])
        self._print_intline("vecopt success", cnt[Counters.OPT_VECTORIZED])
        cpu = self.cpu
        if cpu is not None:   # for some tests
            self._print_intline("Total # of loops",
                                cpu.tracker.total_compiled_loops)
            self._print_intline("Total # of bridges",
                                cpu.tracker.total_compiled_bridges)
            self._print_intline("Freed # of loops",
                                cpu.tracker.total_freed_loops)
            self._print_intline("Freed # of bridges",
                                cpu.tracker.total_freed_bridges)

    def _print_line_time(self, string, i, tim):
        final = "%s:%s\t%d\t%f" % (string, " " * max(0, 13-len(string)), i, tim)
        debug_print(final)

    def _print_intline(self, string, i):
        final = string + ':' + " " * max(0, 16-len(string))
        final += '\t' + str(i)
        debug_print(final)


class BrokenProfilerData(JitException):
    pass



# only for the LLInterpreter.  Don't use directly.

from rpython.rtyper.lltypesystem.lltype import malloc, free, typeOf
from rpython.rtyper.lltypesystem.llmemory import weakref_create, weakref_deref

setfield = setattr
from operator import setitem as setarrayitem
from rpython.rlib.rgc import can_move, collect, enable, disable, isenabled, add_memory_pressure, collect_step

def setinterior(toplevelcontainer, inneraddr, INNERTYPE, newvalue,
                offsets=None):
    assert typeOf(newvalue) == INNERTYPE
    # xxx access the address object's ref() directly for performance
    inneraddr.ref()[0] = newvalue

from rpython.rtyper.lltypesystem.lltype import cast_ptr_to_int as gc_id

def weakref_create_getlazy(objgetter):
    return weakref_create(objgetter())


def shrink_array(p, smallersize):
    return False


def thread_run():
    pass

def thread_start():
    pass

def thread_die():
    pass

def pin(obj):
    return False

def unpin(obj):
    raise AssertionError("pin() always returns False, "
                         "so unpin() should not be called")

def _is_pinned(obj):
    return False
